"""Stub for loading and processing recorded data"""

import argparse
import json

import numpy as np

import acconeer.exptool as et


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename")
    parser.add_argument("--plot", action="store_true")
    args = parser.parse_args()
    filename = args.filename

    record = et.a111.recording.load(filename)

    print_record(record)

    if not args.plot:
        return

    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()

    depths = et.a111.get_range_depths(record.sensor_config, record.session_info)

    if len(record.data.shape) == 3:  # Envelope, IQ, Power bins
        x = np.arange(len(record.data))
        y = depths
        z = np.abs(record.data[:, 0, :]).T

        ax.pcolormesh(x, y, z)
        ax.set_xlabel("Sweep index")
        ax.set_ylabel("Depth (m)")
    else:
        y = np.abs(record.data[:, 0, :, :].mean(axis=1))

        ax.plot(y)
        ax.set_xlabel("Frame index")
        ax.set_ylabel("Mean sweep amplitude")

        ax.legend([f"{d:.2f}" for d in depths], title="Depth")

    fig.tight_layout()
    plt.show()


def print_record(record):
    print("Mode:", record.mode.name.lower())
    print()
    print(record.sensor_config)
    print()
    print("Session info")

    for k, v in record.session_info.items():
        print("  {:.<35} {}".format(k + " ", v))

    print()
    print("Data shape:", record.data.shape)
    print("Data dtype:", record.data.dtype)
    print()
    print("Last data info (first sensor):")

    for k, v in record.data_info[-1][0].items():
        print("  {:.<35} {}".format(k + " ", v))

    ts = record.sample_times
    if ts is not None and ts.size >= 2:
        print()
        mean_dt = (ts[-1] - ts[0]) / (ts.size - 1)
        mean_f = 1 / mean_dt
        print("Mean sample rate (client side): {:.2f} Hz".format(mean_f))

    print("\n")

    print("Module (processing) key:", record.module_key)

    if record.processing_config_dump is None:
        print("No processing config dump")
    else:
        print("Processing config dump")
        for k, v in json.loads(record.processing_config_dump).items():
            print("  {:.<35} {}".format(k + " ", v))

    print("\n")

    m = {
        "RSS version": record.rss_version,
        "acconeer.exptool library version": record.lib_version,
        "Timestamp": record.timestamp,
    }

    for k, v in m.items():
        print("{:.<37} {}".format(k + " ", v))

    if record.note:
        print()
        print("Note: " + str(record.note))


if __name__ == "__main__":
    main()



import abc
import logging

from packaging import version

from acconeer.exptool._structs import configbase
from acconeer.exptool.a111 import SDK_VERSION, _modes


log = logging.getLogger(__name__)


class BaseClient(abc.ABC):
    @abc.abstractmethod
    def __init__(self, **kwargs):
        self._squeeze = kwargs.pop("squeeze", True)

        if kwargs:
            a_key = next(iter(kwargs.keys()))
            raise TypeError("Got unexpected keyword argument ({})".format(a_key))

        self._connected = False
        self._session_setup_done = False
        self._streaming_started = False
        self.supported_modes = None

    def connect(self):
        """Initiates a connection with the device.

        :return: A dict containing information about the device, including SDK version
        :rtype: dict
        """
        if self._connected:
            raise ClientError("already connected")

        info = self._connect()
        self._connected = True

        if info is None:
            info = {}

        if not info.get("mock"):
            try:
                log.info("reported version: {}".format(info["version_str"]))

                if info["strict_version"] < version.parse(SDK_VERSION):
                    log.warning("old server version - please upgrade server")
                elif info["strict_version"] > version.parse(SDK_VERSION):
                    log.warning("new server version - please upgrade client")
            except KeyError:
                log.warning("could not read software version (might be too old)")

        self.supported_modes = self._get_supported_modes()

        return info

    def setup_session(self, config, check_config=True):
        """
        Sets up a session with the given config.
        Will call ``connect()`` if not already connected.

        :param config: The configuration to use when setting up the session
        :type config: class:`acconeer.exptool.configs`
        :param check_config: If `True` the configuration is checked for errors,
                            defaults to `True`
        :type check_config: bool

        :return: A dict with metadata for the configured session
        :rtype: dict
        """
        if check_config:
            self._check_config(config)

        if self._streaming_started:
            raise ClientError("can't setup session while streaming")

        if not self._connected:
            self.connect()

        if check_config and config.mode not in self.supported_modes:
            raise ClientError("Unsupported mode")

        session_info = self._setup_session(config)
        self._session_setup_done = True
        return session_info

    def start_session(self, config=None, check_config=True):
        """
        Starts the session if previously set up with ``setup_session()``.
        If `config` is provided, ``setup_session()`` will be called.

        :param config: The configuration to use when setting up the session, defaults to `None`
        :type config: class:`acconeer.exptool.configs`, optional
        :param check_config: If `True` the configuration is checked for errors,
                            defaults to `True`
        :type check_config: bool

        :return: If `config` is provided, returns a dict with metadata for the configured session.
                Otherwise, returns `None`
        :rtype: dict or None
        """
        if self._streaming_started:
            raise ClientError("already streaming")

        if config is None:
            ret = None
        else:
            ret = self.setup_session(config, check_config=check_config)

        if not self._session_setup_done:
            raise ClientError("session needs to be set up before starting stream")

        self._start_session()
        self._streaming_started = True
        return ret

    def get_next(self):
        """
        Retrieves the next result. Will block until the result is received.

        :return: A tuple with the result info and data.
                The data shape and type differs between services.

                | **Power Bins:**
                | Shape: (number of sensors, bin count)
                | Type: float64

                | **Envelope:**
                | Shape: (number of sensors, data length)
                | Type: float64

                | **IQ:**
                | Shape: (number of sensors, data length)
                | Type: complex128

                | **Sparse:**
                | Shape: (number of sensors, number of sweeps, number of dephts)
                | Type: float64

                `Number of sensors`, `bin count` and `number of sweeps` can be explicitly set.
                `Data length` and `number of dephts` depend on multiple configuration settings.

                The client takes a parameter ``squeeze``, if set to `True` the first
                dimension (`number of sensors`) is removed when using a single sensor.
                As default ``squeeze`` is  `True`.

        :rtype: tuple[union[list, dict], np.ndarray]
        """
        if not self._streaming_started:
            raise ClientError("must be streaming to get next")

        return self._get_next()

    def stop_session(self):
        """
        Stops the session. All buffered/waiting data is thrown away.
        This function will block until the server has confirmed that the session has ended.
        """
        if not self._streaming_started:
            raise ClientError("not streaming")

        self._stop_session()
        self._streaming_started = False

    def disconnect(self):
        """
        Disconnects the client. ``disconnect()`` will call ``stop_session()``
        if a session is started.
        """
        if not self._connected:
            raise ClientError("not connected")

        if self._streaming_started:
            self.stop_session()

        self._disconnect()
        self._connected = False
        self.supported_modes = None

    def _check_config(self, config):
        try:
            alerts = config.check()
        except AttributeError:
            return

        try:
            error_alert = next(a for a in alerts if a.severity == configbase.Severity.ERROR)
        except StopIteration:
            return

        msg = "error in config: {}: {}".format(error_alert.param, error_alert.msg)
        raise IllegalConfigError(msg)

    @property
    @abc.abstractmethod
    def description(self):
        pass

    def _get_supported_modes(self):
        return set(_modes.Mode)

    @abc.abstractmethod
    def _connect(self):
        pass

    @abc.abstractmethod
    def _setup_session(self, config):
        pass

    @abc.abstractmethod
    def _start_session(self):
        pass

    @abc.abstractmethod
    def _get_next(self):
        pass

    @abc.abstractmethod
    def _stop_session(self):
        pass

    @abc.abstractmethod
    def _disconnect(self):
        pass

    @property
    def squeeze(self):
        return self._squeeze

    @squeeze.setter
    def squeeze(self, squeeze):
        self._squeeze = squeeze


class ClientError(Exception):
    pass


class IllegalConfigError(ClientError):
    pass


class SessionSetupError(ClientError):
    pass


def decode_version_str(version_str: str) -> dict:
    if "-" in version_str:
        strict_version = version.parse(version_str.split("-")[0])
    else:
        strict_version = version.parse(version_str)

    return {
        "version_str": version_str,
        "strict_version": strict_version,
    }



import numpy as np

import acconeer.exptool as et

from .constants import HISTORY_LENGTH_S


def get_sensor_config():
    config = et.a111.EnvelopeServiceConfig()
    config.profile = et.a111.EnvelopeServiceConfig.Profile.PROFILE_1
    config.range_interval = [0.04, 0.05]
    config.running_average_factor = 0.01
    config.maximize_signal_attenuation = True
    config.update_rate = 60
    config.gain = 0.2
    config.repetition_mode = et.a111.EnvelopeServiceConfig.RepetitionMode.SENSOR_DRIVEN
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 2

    signal_tc_s = et.configbase.FloatParameter(
        label="Signal time constant",
        unit="s",
        default_value=5.0,
        limits=(0.01, 10),
        logscale=True,
        updateable=True,
        order=20,
        help="Time constant of the low pass filter for the signal.",
    )

    rel_dev_tc_s = et.configbase.FloatParameter(
        label="Relative deviation time constant",
        unit="s",
        default_value=0.2,
        limits=(0.01, 2),
        logscale=True,
        updateable=True,
        order=30,
        help="Time constant of the low pass filter for the relative deviation.",
    )

    threshold = et.configbase.FloatParameter(
        label="Detection threshold",
        default_value=0.2,
        decimals=3,
        limits=(0.001, 0.5),
        updateable=True,
        logscale=True,
        order=10,
        help='Level at which the detector output is considered as a "button press". '
        "Note that this might need adjustment depending "
        "on different board models in order to detect movement.",
    )

    buttonpress_length_s = et.configbase.FloatParameter(
        label="Button press length",
        unit="s",
        default_value=2.0,
        limits=(0.01, 5),
        logscale=False,
        updateable=True,
        order=40,
        help="The time after a detected button press when no further detection should occur.",
    )

    def check_sensor_config(self, sensor_config):
        alerts = {
            "processing": [],
            "sensor": [],
        }

        alerts["processing"].append(
            et.configbase.Info(
                "threshold", "Threshold level should be adjusted \ndepending on board model."
            )
        )

        return alerts


class Processor:
    # lp(f): low pass (filtered)
    # cut: cutoff frequency [Hz]
    # tc: time constant [s]
    # sf: smoothing factor [dimensionless]

    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        assert sensor_config.update_rate is not None

        self.f = sensor_config.update_rate

        self.signal_history = np.zeros(int(round(self.f * HISTORY_LENGTH_S)))
        self.signal_lp_history = np.zeros(int(round(self.f * HISTORY_LENGTH_S)))
        self.rel_dev_history = np.zeros(int(round(self.f * HISTORY_LENGTH_S)))
        self.rel_dev_lp_history = np.zeros(int(round(self.f * HISTORY_LENGTH_S)))
        self.detection_history = []
        self.signal_lp = 0.0
        self.rel_dev_lp = 0
        self.sweep_index = 0
        self.last_detection_sweep = 0

        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        self.threshold = processing_config.threshold

        self.sf_signal = np.exp(-1.0 / (processing_config.signal_tc_s * self.f))
        self.sf_rel_dev = np.exp(-1.0 / (processing_config.rel_dev_tc_s * self.f))
        self.buttonpress_length_sweeps = processing_config.buttonpress_length_s * self.f

    def process(self, data, data_info):
        sweep = data

        # Sum the full sweep to a single number
        signal = np.mean(sweep)

        # Exponential filtering of the signal
        sf = min(self.sf_signal, 1.0 - 1.0 / (1.0 + self.sweep_index))
        self.signal_lp = sf * self.signal_lp + (1.0 - sf) * signal

        # The relative difference
        rel_dev = np.square((signal - self.signal_lp) / self.signal_lp)

        # Exponential filtering of the difference
        sf = min(self.sf_rel_dev, 1.0 - 1.0 / (1.0 + self.sweep_index))
        self.rel_dev_lp = sf * self.rel_dev_lp + (1.0 - sf) * rel_dev

        # Check detection
        detection = False
        sweeps_since_last_detect = self.sweep_index - self.last_detection_sweep
        detection_long_enough_ago = sweeps_since_last_detect > self.buttonpress_length_sweeps
        over_threshold = self.rel_dev_lp > self.threshold
        if over_threshold and detection_long_enough_ago:
            self.last_detection_sweep = self.sweep_index
            detection = True

        # Save all signal in history arrays.
        self.signal_history = np.roll(self.signal_history, -1)
        self.signal_history[-1] = signal

        self.signal_lp_history = np.roll(self.signal_lp_history, -1)
        self.signal_lp_history[-1] = self.signal_lp

        self.rel_dev_history = np.roll(self.rel_dev_history, -1)
        self.rel_dev_history[-1] = rel_dev

        self.rel_dev_lp_history = np.roll(self.rel_dev_lp_history, -1)
        self.rel_dev_lp_history[-1] = self.rel_dev_lp

        if detection:
            self.detection_history.append(self.sweep_index)

        while (
            len(self.detection_history) > 0
            and self.sweep_index - self.detection_history[0] > HISTORY_LENGTH_S * self.f
        ):
            self.detection_history.remove(self.detection_history[0])

        out_data = {
            "signal_history": self.signal_history,
            "signal_lp_history": self.signal_lp_history,
            "rel_dev_history": self.rel_dev_history,
            "rel_dev_lp_history": self.rel_dev_lp_history,
            "detection_history": (np.array(self.detection_history) - self.sweep_index) / self.f,
            "detection": detection,
            "sweep_index": self.sweep_index,
            "threshold": self.threshold,
        }

        self.sweep_index += 1

        return out_data



from copy import copy
from enum import Enum
from typing import Optional

import numpy as np

import acconeer.exptool as et

from .calibration import DistanceDetectorCalibration


PEAK_MERGE_LIMIT_M = 0.005


def get_sensor_config():
    config = et.a111.EnvelopeServiceConfig()
    config.range_interval = [0.2, 0.6]
    config.update_rate = 40
    config.gain = 0.5
    config.running_average_factor = 0  # Use averaging in detector instead of in API

    return config


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.session_info = session_info

        self.f = sensor_config.update_rate

        self.num_depths = self.session_info["data_length"]

        self.current_mean_sweep = np.zeros(self.num_depths)
        self.last_mean_sweep = np.full(self.num_depths, np.nan)
        self.sweeps_since_mean = 0

        self.sc_sum_bg_sweeps = np.zeros(self.num_depths)
        self.sc_sum_squared_bg_sweeps = np.zeros(self.num_depths)
        self.sc_used_threshold = np.full(self.num_depths, np.nan)
        self.calibration: Optional[DistanceDetectorCalibration] = None

        self.sc_bg_calculated = False

        self.history_length_s = processing_config.history_length_s
        self.main_peak_hist_sweep_idx = []
        self.main_peak_hist_dist = []
        self.minor_peaks_hist_sweep_idx = []
        self.minor_peaks_hist_dist = []
        self.above_thres_hist_sweep_idx = []
        self.above_thres_hist_dist = []

        self.r = et.a111.get_range_depths(sensor_config, session_info)
        self.dr = self.r[1] - self.r[0]
        self.sweep_index = 0

        self.update_processing_config(processing_config)

        self.update_calibration(calibration)

    def update_processing_config(self, processing_config):
        self.nbr_average = processing_config.nbr_average
        self.threshold_type = processing_config.threshold_type
        self.peak_sorting_method = processing_config.peak_sorting_type

        self.fixed_threshold_level = processing_config.fixed_threshold

        self.sc_sensitivity = processing_config.sc_sensitivity
        self.sc_bg_nbr_sweeps = processing_config.sc_nbr_sweep_for_bg

        self.idx_cfar_pts = np.round(
            (
                processing_config.cfar_guard_m / 2.0 / self.dr
                + np.arange(processing_config.cfar_window_m / self.dr)
            )
        )

        self.cfar_one_sided = processing_config.cfar_one_sided
        self.cfar_sensitivity = processing_config.cfar_sensitivity

        self.update_sc_threshold()
        self.max_nbr_peaks_to_plot = processing_config.max_nbr_peaks_to_plot
        self.history_length_s = processing_config.history_length_s

    def update_calibration(self, new_calibration: DistanceDetectorCalibration):
        self.calibration = new_calibration
        self.update_sc_threshold()

    @property
    def sc_used_mean(self):
        if self.calibration is None:
            return np.full(self.num_depths, np.nan)
        else:
            return self.calibration.stationary_clutter_mean

    @property
    def sc_used_std(self):
        if self.calibration is None:
            return np.full(self.num_depths, np.nan)
        else:
            return self.calibration.stationary_clutter_std

    def update_sc_threshold(self):
        self.sc_used_threshold = (
            self.sc_used_mean + (1.0 / (self.sc_sensitivity + 1e-10) - 1.0) * self.sc_used_std
        )

    def get_sc_threshold(self, sweep) -> Optional[DistanceDetectorCalibration]:
        # Collect first sweeps to construct a stationary clutter threshold
        # Accumulate sweeps instead of saving each for lower memory footprint
        calibration = None

        if self.sweep_index < self.sc_bg_nbr_sweeps:
            self.sc_sum_bg_sweeps += sweep
            self.sc_sum_squared_bg_sweeps += np.square(sweep)

        if self.sweep_index >= self.sc_bg_nbr_sweeps - 1 and not self.sc_bg_calculated:
            sc_bg_sweep_mean = self.sc_sum_bg_sweeps / self.sc_bg_nbr_sweeps
            mean_square = self.sc_sum_squared_bg_sweeps / self.sc_bg_nbr_sweeps
            square_mean = np.square(sc_bg_sweep_mean)
            sc_bg_sweep_std = np.sqrt(
                (mean_square - square_mean) * self.sc_bg_nbr_sweeps / (self.sc_bg_nbr_sweeps - 1)
            )
            calibration = DistanceDetectorCalibration(
                sc_bg_sweep_mean,
                sc_bg_sweep_std,
            )

            self.sc_bg_calculated = True

        return calibration

    def calculate_cfar_threshold(self, sweep, idx_cfar_pts, alpha, one_side):

        threshold = np.full(sweep.shape, np.nan)

        start_idx = np.max(idx_cfar_pts)
        if one_side:
            rel_indexes = -idx_cfar_pts
            end_idx = sweep.size
        else:
            rel_indexes = np.concatenate((-idx_cfar_pts, +idx_cfar_pts), axis=0)
            end_idx = sweep.size - start_idx

        for idx in np.arange(start_idx, end_idx):
            threshold[int(idx)] = (
                1.0 / (alpha + 1e-10) * np.mean(sweep[(idx + rel_indexes).astype(int)])
            )

        return threshold

    def find_first_point_above_threshold(self, sweep, threshold):

        if threshold is None or np.all(np.isnan(threshold)):
            return None

        points_above = sweep > threshold

        if not np.any(points_above):
            return None

        return np.argmax(points_above)

    def find_peaks(self, sweep, threshold):
        #  Not written for optimal speed.

        if threshold is None or np.all(np.isnan(threshold)):
            return []

        found_peaks = []

        # Note: at least 3 samples above threshold are required to form a peak

        d = 1
        N = len(sweep)
        while d < (N - 1):
            # Skip to when threshold starts, applicable only for CFAR
            if np.isnan(threshold[d - 1]):
                d += 1
                continue

            # Break when threshold ends, applicable only for CFAR
            if np.isnan(threshold[d + 1]):
                break

            # At this point, threshold is defined (not Nan)

            # If the current point is not over threshold, the next will not be a peak
            if sweep[d] <= threshold[d]:
                d += 2
                continue

            # Continue if previous point is not over threshold
            if sweep[d - 1] <= threshold[d - 1]:
                d += 1
                continue

            # Continue if this point isn't larger than the previous
            if sweep[d - 1] >= sweep[d]:
                d += 1
                continue

            # A peak is either a single point or a plateau consisting of several equal points,
            # all over their threshold. The closest neighboring points on each side of the
            # point/plateau must have a lower value and be over their threshold.
            # Now, decide if the following point(s) are a peak:

            d_upper = d + 1
            while True:
                if (d_upper) >= (N - 1):  # If out of range or on last point
                    break

                if np.isnan(threshold[d_upper]):
                    break

                if sweep[d_upper] <= threshold[d_upper]:
                    break

                if sweep[d_upper] > sweep[d]:
                    break
                elif sweep[d_upper] < sweep[d]:
                    delta = d_upper - d
                    found_peaks.append(d + int(np.ceil((delta - 1) / 2.0)))
                    break
                else:  # equal
                    d_upper += 1

            d = d_upper

        return found_peaks

    def merge_peaks(self, peak_indexes, merge_max_range):
        merged_peaks = copy(peak_indexes)

        while True:
            num_neighbors = np.zeros(len(merged_peaks))  # number of neighbors
            for i, p in enumerate(merged_peaks):
                num_neighbors[i] = np.sum(np.abs(np.array(merged_peaks) - p) < merge_max_range)

            # First peak with max number of neighbors
            i_peak = np.argmax(num_neighbors)  # returns arg of first max

            if num_neighbors[i_peak] <= 1:
                break

            peak = merged_peaks[i_peak]

            remove_mask = np.abs(np.array(merged_peaks) - peak) < merge_max_range
            peaks_to_remove = np.array(merged_peaks)[remove_mask]

            for p in peaks_to_remove:
                merged_peaks.remove(p)

            # Add back mean peak
            merged_peaks.append(int(round(np.mean(peaks_to_remove))))

            merged_peaks.sort()

        return merged_peaks

    def sort_peaks(self, peak_indexes, sweep):
        amp = np.array([sweep[int(i)] for i in peak_indexes])
        r = np.array([self.r[int(i)] for i in peak_indexes])

        PeakSorting = ProcessingConfiguration.PeakSorting
        if self.peak_sorting_method == PeakSorting.CLOSEST:
            quantity_to_sort = r
        elif self.peak_sorting_method == PeakSorting.STRONGEST:
            quantity_to_sort = -amp
        elif self.peak_sorting_method == PeakSorting.STRONGEST_REFLECTOR:
            quantity_to_sort = -amp * r**2
        elif self.peak_sorting_method == PeakSorting.STRONGEST_FLAT_REFLECTOR:
            quantity_to_sort = -amp * r
        else:
            raise Exception("Unknown peak sorting method")

        return [peak_indexes[i] for i in quantity_to_sort.argsort()]

    def process(self, data, data_info):
        sweep = data

        # Accumulate sweeps for stationary clutter threshold and check if user has
        # loaded one from disk
        new_calibration = self.get_sc_threshold(sweep)

        # Average envelope sweeps, written to handle varying nbr_average
        weight = 1.0 / (1.0 + self.sweeps_since_mean)
        self.current_mean_sweep = weight * sweep + (1.0 - weight) * self.current_mean_sweep
        self.sweeps_since_mean += 1

        # Determining threshold
        if self.threshold_type is ProcessingConfiguration.ThresholdType.FIXED:
            threshold = self.fixed_threshold_level * np.ones(sweep.size)
        elif self.threshold_type is ProcessingConfiguration.ThresholdType.RECORDED:
            threshold = self.sc_used_threshold
        elif self.threshold_type is ProcessingConfiguration.ThresholdType.CFAR:
            threshold = self.calculate_cfar_threshold(
                self.current_mean_sweep,
                self.idx_cfar_pts,
                self.cfar_sensitivity,
                self.cfar_one_sided,
            )
        else:
            print("Unknown thresholding method")

        found_peaks = None

        # If a new averaged sweep is ready for processing
        if self.sweeps_since_mean >= self.nbr_average:
            self.sweeps_since_mean = 0
            self.last_mean_sweep = self.current_mean_sweep.copy()
            self.current_mean_sweep *= 0

            # Find the first delay over threshold. Used in tank-level when monitoring changes
            # in the direct leakage.
            first_point_above_threshold = self.find_first_point_above_threshold(
                self.last_mean_sweep, threshold
            )

            # First peak-finding, then peak-merging, finallay peak sorting.
            found_peaks = self.find_peaks(self.last_mean_sweep, threshold)
            if len(found_peaks) > 1:
                found_peaks = self.merge_peaks(found_peaks, np.round(PEAK_MERGE_LIMIT_M / self.dr))
                found_peaks = self.sort_peaks(found_peaks, self.last_mean_sweep)

            # Adding main peak to history
            if len(found_peaks) > 0:
                self.main_peak_hist_sweep_idx.append(self.sweep_index)
                self.main_peak_hist_dist.append(self.r[found_peaks[0]])

            # Adding minor peaks to history
            for i in range(1, min(len(found_peaks), self.max_nbr_peaks_to_plot)):
                self.minor_peaks_hist_sweep_idx.append(self.sweep_index)
                self.minor_peaks_hist_dist.append(self.r[found_peaks[i]])

            # Adding first distance above threshold to history
            if first_point_above_threshold is not None:
                self.above_thres_hist_sweep_idx.append(self.sweep_index)
                self.above_thres_hist_dist.append(self.r[first_point_above_threshold])

            # Removing old main peaks from history
            while (
                len(self.main_peak_hist_sweep_idx) > 0
                and self.sweep_index - self.main_peak_hist_sweep_idx[0]
                > self.history_length_s * self.f
            ):
                self.main_peak_hist_sweep_idx.pop(0)
                self.main_peak_hist_dist.pop(0)

            # Removing old minor peaks from history
            while (
                len(self.minor_peaks_hist_sweep_idx) > 0
                and self.sweep_index - self.minor_peaks_hist_sweep_idx[0]
                > self.history_length_s * self.f
            ):
                self.minor_peaks_hist_sweep_idx.pop(0)
                self.minor_peaks_hist_dist.pop(0)

            # Removing old first distance above threshold from history
            while (
                len(self.above_thres_hist_sweep_idx) > 0
                and self.sweep_index - self.above_thres_hist_sweep_idx[0]
                > self.history_length_s * self.f
            ):
                self.above_thres_hist_sweep_idx.pop(0)
                self.above_thres_hist_dist.pop(0)

        out_data = {
            "sweep": sweep,
            "last_mean_sweep": self.last_mean_sweep,
            "threshold": threshold,
            "main_peak_hist_sweep_s": (
                (np.array(self.main_peak_hist_sweep_idx) - self.sweep_index) / self.f
            ),
            "main_peak_hist_dist": np.array(self.main_peak_hist_dist),
            "minor_peaks_hist_sweep_s": (
                (np.array(self.minor_peaks_hist_sweep_idx) - self.sweep_index) / self.f
            ),
            "minor_peaks_hist_dist": np.array(self.minor_peaks_hist_dist),
            "above_thres_hist_sweep_s": (
                (np.array(self.above_thres_hist_sweep_idx) - self.sweep_index) / self.f
            ),
            "above_thres_hist_dist": np.array(self.above_thres_hist_dist),
            "sweep_index": self.sweep_index,
            "found_peaks": found_peaks,
        }

        if new_calibration:
            out_data["new_calibration"] = new_calibration

        self.sweep_index += 1

        return out_data


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    class ThresholdType(Enum):
        FIXED = "Fixed"
        RECORDED = "Recorded"
        CFAR = "CFAR"

    class PeakSorting(Enum):
        STRONGEST = "Strongest signal"
        CLOSEST = "Closest signal"
        STRONGEST_REFLECTOR = "Strongest reflector"
        STRONGEST_FLAT_REFLECTOR = "Strongest flat reflector"

    VERSION = 2

    nbr_average = et.configbase.FloatParameter(
        label="Sweep averaging",
        default_value=5,
        limits=(1, 100),
        logscale=True,
        decimals=0,
        updateable=True,
        order=0,
        visible=True,
        help=(
            "The number of envelope sweeps to be average into one then used for"
            " distance detection."
        ),
    )

    threshold_type = et.configbase.EnumParameter(
        label="Threshold type",
        default_value=ThresholdType.FIXED,
        enum=ThresholdType,
        updateable=True,
        order=5,
        help="Setting the type of threshold",
    )

    fixed_threshold = et.configbase.FloatParameter(
        label="Fixed threshold level",
        default_value=800,
        limits=(1, 20000),
        decimals=0,
        updateable=True,
        order=10,
        visible=lambda conf: conf.threshold_type == conf.ThresholdType.FIXED,
        help=(
            "Sets the value of fixed threshold. The threshold has this constant value over"
            " the full sweep."
        ),
    )

    sc_nbr_sweep_for_bg = et.configbase.FloatParameter(
        label="Number of sweeps for background estimation",
        default_value=20,
        limits=(2, 200),
        decimals=0,
        visible=lambda conf: conf.threshold_type == conf.ThresholdType.RECORDED,
        updateable=True,
        order=20,
        help=(
            "The number of (non-averaged) sweeps collected for calculating the Stationary"
            " Clutter threshold."
        ),
    )

    sc_sensitivity = et.configbase.FloatParameter(
        label="Stationary clutter sensitivity",
        default_value=0.3,
        limits=(0.01, 1),
        logscale=True,
        visible=lambda conf: conf.threshold_type == conf.ThresholdType.RECORDED,
        decimals=4,
        updateable=True,
        order=24,
        help=(
            "Value between 0 and 1 that sets the threshold. A low sensitivity will set a "
            "high threshold, resulting in only few false alarms but might result in "
            "missed detections."
        ),
    )

    cfar_sensitivity = et.configbase.FloatParameter(
        label="CFAR sensitivity",
        default_value=0.5,
        limits=(0.01, 1),
        logscale=True,
        visible=lambda conf: conf.threshold_type == conf.ThresholdType.CFAR,
        decimals=4,
        updateable=True,
        order=40,
        help=(
            "Value between 0 and 1 that sets the threshold. A low sensitivity will set a "
            "high threshold, resulting in only few false alarms but might result in "
            "missed detections."
        ),
    )

    cfar_guard_m = et.configbase.FloatParameter(
        label="CFAR guard",
        default_value=0.12,
        limits=(0.01, 0.2),
        unit="m",
        decimals=3,
        visible=lambda conf: conf.threshold_type == conf.ThresholdType.CFAR,
        updateable=True,
        order=41,
        help=(
            "Range around the distance of interest that is omitted when calculating "
            "CFAR threshold. Can be low, ~40 mm, for Profile 1, and should be "
            "increased for higher Profiles."
        ),
    )

    cfar_window_m = et.configbase.FloatParameter(
        label="CFAR window",
        default_value=0.03,
        limits=(0.001, 0.2),
        unit="m",
        decimals=3,
        visible=lambda conf: conf.threshold_type == conf.ThresholdType.CFAR,
        updateable=True,
        order=42,
        help="Range next to the CFAR guard from which the threshold level will be calculated.",
    )

    cfar_one_sided = et.configbase.BoolParameter(
        label="Use only lower distance to set threshold",
        default_value=False,
        visible=lambda conf: conf.threshold_type == conf.ThresholdType.CFAR,
        updateable=True,
        order=43,
        help=(
            "Instead of determining the CFAR threshold from sweep amplitudes from "
            "distances both closer and a farther, use only closer. Helpful e.g. for "
            "fluid level in small tanks, where many multipath signal can apprear "
            "just after the main peak."
        ),
    )

    peak_sorting_type = et.configbase.EnumParameter(
        label="Peak sorting",
        default_value=PeakSorting.STRONGEST,
        enum=PeakSorting,
        updateable=True,
        order=100,
        help="Setting the type of peak sorting method.",
    )

    history_length_s = et.configbase.FloatParameter(
        default_value=10,
        limits=(3, 1000),
        updateable=True,
        logscale=True,
        unit="s",
        label="History length",
        order=198,
        help="Length of time history for plotting.",
    )

    max_nbr_peaks_to_plot = et.configbase.IntParameter(
        default_value=1,
        limits=(1, 15),
        updateable=True,
        unit="peaks",
        label="Peaks per sweep in plot",
        order=200,
        help="The maximum number of peaks per averaged sweep to be plotted in the lower figure.",
    )

    show_first_above_threshold = et.configbase.BoolParameter(
        label="Show first distance above threshold",
        default_value=False,
        updateable=True,
        order=201,
        help=(
            "When detect in the presence of object very close to the sensor, the "
            "strong direct leakage might cause that no well shaped peaks are detected, "
            "even though the envelope signal is above the threshold. Therefore the "
            "first distace where the signal is above the threshold can be used as an "
            "alternative to peak detection."
        ),
    )

    def check_sensor_config(self, sensor_config):
        alerts = {
            "processing": [],
            "sensor": [],
        }
        if sensor_config.update_rate is None:
            alerts["sensor"].append(et.configbase.Error("update_rate", "Must be set"))

        if not sensor_config.noise_level_normalization:
            if self.threshold_type == self.ThresholdType.FIXED:
                alerts["sensor"].append(
                    et.configbase.Warning(
                        "noise_level_normalization",
                        (
                            "Enabling noise level normalization is\n"
                            "recommended with Fixed threshold"
                        ),
                    )
                )

        return alerts



from acconeer.exptool.a111.algo import ModuleFamily, ModuleInfo
from acconeer.exptool.a111.algo.utils import PassthroughProcessor

from ._processor import get_sensor_config
from .ui import PGUpdater


module_info = ModuleInfo(
    key="power_bins",
    label="Power bins",
    pg_updater=PGUpdater,
    processing_config_class=lambda: None,
    module_family=ModuleFamily.SERVICE,
    sensor_config_class=get_sensor_config,
    processor=PassthroughProcessor,
    multi_sensor=False,
    docs_url="https://docs.acconeer.com/en/latest/services/pb.html",
)



import acconeer.exptool as et


def main():
    args = et.a111.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = et.a111.Client(**et.a111.get_client_args(args))

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = args.sensors

    print(config)

    connect_info = client.connect()
    print("connect info:")
    print_dict(connect_info)

    session_info = client.start_session(config)
    print("session_info:")
    print_dict(session_info)

    data_info, data = client.get_next()
    print("data_info:")
    print_dict(data_info)

    client.disconnect()


def print_dict(d):
    for k, v in d.items():
        print("  {:.<35} {}".format(k + " ", v))


if __name__ == "__main__":
    main()



import shutil
import zipfile
from pathlib import Path

import requests


def main():
    here = Path(__file__).resolve().parent
    package_dir = here / "package"
    package_zip = here / "portable.zip"

    if package_zip.exists():
        return

    path = package_dir / "tools" / "get-pip.py"
    if not path.exists():
        download("https://bootstrap.pypa.io/get-pip.py", path)

    url = "https://www.python.org/ftp/python/3.9.10/python-3.9.10-embed-amd64.zip"
    py_zip_name = url.split("/")[-1]
    py_dir_name = str(Path(py_zip_name).stem)
    py_target_dir = package_dir / "tools" / py_dir_name
    if not py_target_dir.exists():
        py_zip_path = here / py_zip_name
        if not py_zip_path.exists():
            download(url, py_zip_path)

        with zipfile.ZipFile(py_zip_path, "r") as z:
            z.extractall(py_target_dir)

        pth_file = py_target_dir / "python39._pth"

        with open(pth_file, "r") as f:
            contents = f.read()

        if "#import site" in contents:
            with open(pth_file, "w") as f:
                f.write(contents.replace("#import site", "import site"))

    shutil.make_archive(package_zip.with_suffix(""), "zip", package_dir)


def download(url, filename):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)


if __name__ == "__main__":
    main()



import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.r = et.a111.get_range_depths(sensor_config, session_info)

        self.setup_is_done = False

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        # Hide the first_distance_above_threshold data
        self.first_distance_above_threshold.setVisible(
            processing_config.show_first_above_threshold
        )

        # ...and hide the marker and text in the legend.
        self.hist_plot.legend.items[2][0].setVisible(processing_config.show_first_above_threshold)
        self.hist_plot.legend.items[2][1].setVisible(processing_config.show_first_above_threshold)

        self.hist_plot.setXRange(-processing_config.history_length_s, 0)

    def setup(self, win):
        win.setWindowTitle("Acconeer Distance Detector")

        # Sweep Plot
        self.sweep_plot = win.addPlot(title="Sweep and threshold")
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.setMouseEnabled(x=False, y=False)
        self.sweep_plot.hideButtons()
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("bottom", "Distance (mm)")
        self.sweep_plot.setYRange(0, 20000)
        self.sweep_plot.setXRange(1000.0 * self.r[0], 1000.0 * self.r[-1])

        self.sweep_curve = self.sweep_plot.plot(
            pen=et.utils.pg_pen_cycler(5),
            name="Envelope sweep",
        )

        self.mean_sweep_curve = self.sweep_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=3),
            name="Mean Envelope sweep",
        )

        self.threshold_curve = self.sweep_plot.plot(
            pen=et.utils.pg_pen_cycler(1),
            name="Threshold",
        )

        self.smooth_max_sweep = et.utils.SmoothMax(
            self.sensor_config.update_rate,
            hysteresis=0.6,
            tau_decay=3,
        )

        self.peak_lines = []
        for i in range(3):
            color_idx = 1 if i > 0 else 0
            width = 2 if i == 0 else 1
            color_tuple = et.utils.hex_to_rgb_tuple(et.utils.color_cycler(color_idx))
            line = pg.InfiniteLine(pen=pg.mkPen(pg.mkColor(*color_tuple, 150), width=width))
            self.sweep_plot.addItem(line)
            self.peak_lines.append(line)

        self.peak_text = pg.TextItem(
            anchor=(0, 1),
            color=et.utils.color_cycler(0),
            fill=pg.mkColor(0xFF, 0xFF, 0xFF, 150),
        )
        self.peak_text.setPos(self.r[0] * 1000, 0)
        self.peak_text.setZValue(100)
        self.sweep_plot.addItem(self.peak_text)

        win.nextRow()

        # Detection history Plot
        self.hist_plot = win.addPlot(title="Detected peaks")
        self.hist_plot.setMenuEnabled(False)
        self.hist_plot.setMouseEnabled(x=False, y=False)
        self.hist_plot.hideButtons()
        self.hist_plot.showGrid(x=True, y=True)
        self.hist_plot.addLegend()
        self.hist_plot.setLabel("bottom", "Time history (s)")
        self.hist_plot.setLabel("left", "Distance (mm)")
        self.hist_plot.setXRange(-10, 0)
        self.hist_plot.setYRange(1000.0 * self.r[0], 1000.0 * self.r[-1])

        self.main_peak = self.hist_plot.plot(
            pen=None,
            symbol="o",
            symbolSize=8,
            symbolPen="k",
            symbolBrush=et.utils.color_cycler(0),
            name="Main peak",
        )

        self.minor_peaks = self.hist_plot.plot(
            pen=None,
            symbol="o",
            symbolSize=5,
            symbolPen="k",
            symbolBrush=et.utils.color_cycler(1),
            name="Minor peaks",
        )

        self.first_distance_above_threshold = self.hist_plot.plot(
            pen=None,
            symbol="o",
            symbolSize=3,
            symbolPen="k",
            symbolBrush=et.utils.color_cycler(2),
            name="First distance above threshold",
            visible=False,
        )

        self.setup_is_done = True
        self.update_processing_config()

    def update(self, data):
        self.sweep_curve.setData(1000.0 * self.r, data["sweep"])
        self.mean_sweep_curve.setData(1000.0 * self.r, data["last_mean_sweep"])
        self.threshold_curve.setData(1000.0 * self.r, data["threshold"])

        m = np.nanmax(
            np.concatenate(
                [
                    2 * data["threshold"],
                    data["sweep"],
                    data["last_mean_sweep"],
                ]
            )
        )
        ymax = self.smooth_max_sweep.update(m)
        self.sweep_plot.setYRange(0, ymax)

        self.main_peak.setData(data["main_peak_hist_sweep_s"], 1000 * data["main_peak_hist_dist"])
        self.minor_peaks.setData(
            data["minor_peaks_hist_sweep_s"], 1000 * data["minor_peaks_hist_dist"]
        )
        self.first_distance_above_threshold.setData(
            data["above_thres_hist_sweep_s"], 1000 * data["above_thres_hist_dist"]
        )

        if data["found_peaks"] is not None:
            peaks = np.take(self.r, data["found_peaks"]) * 1000.0
            for i, line in enumerate(self.peak_lines):
                try:
                    peak = peaks[i]
                except (TypeError, IndexError):
                    line.hide()
                else:
                    line.setPos(peak)
                    line.show()

            if data["found_peaks"]:
                amplitude = data["sweep"][data["found_peaks"][0]]
                text = "{:.2f} mm, {:.1f}".format(peaks[0], amplitude)
            else:
                text = "-"

            self.peak_text.setText(text)



from typing import Any, List, Optional, Tuple

import acconeer.exptool as et

from .base import BaseClient
from .common import Link, LinkArg, Protocol, ProtocolArg
from .json.client import SocketClient
from .reg.client import SPIClient, UARTClient


class ClientFactory:
    """
    A Factory that sole responsibility is the creation of `BaseClient` subclasses and
    Error handling.
    """

    @classmethod
    def from_kwargs(
        cls,
        *,
        protocol: Optional[ProtocolArg] = None,
        link: Optional[LinkArg] = None,
        **kwargs: Any,
    ) -> BaseClient:
        cls._check_conflicting_kwargs(protocol=protocol, link=link, **kwargs)

        if link is None:
            link = cls._try_infer_link_from_kwargs(**kwargs)

        if protocol is None:
            protocol = cls._try_infer_protocol_from_kwargs(**kwargs)

        if link == Link.UART:
            port = cls._handle_passed_serial_port(kwargs.pop("serial_port", None))
            return cls._try_get_serial_client(port, protocol, **kwargs)

        if link == Link.SOCKET:
            host = cls._handle_passed_host(kwargs.pop("host", None))
            return cls._try_get_socket_client(host, protocol, **kwargs)

        if link == Link.SPI:
            return SPIClient(**kwargs)

        raise ValueError(f"Could not determine client type. Unrecognized link: {link}")

    @classmethod
    def _check_conflicting_kwargs(cls, **kwargs):
        CONFLICTING_KEY_COMBINATIONS = [{"host", "serial_port"}]

        keys_to_check = {k for k, v in kwargs.items() if v is not None}

        for key_combination in CONFLICTING_KEY_COMBINATIONS:
            # "<=" operator between sets is a subset-check.
            if key_combination <= keys_to_check:
                raise ValueError(
                    f"Incompatible keyword-arguments: {key_combination}. "
                    + "These cannot be used together."
                )

    @classmethod
    def _try_infer_link_from_kwargs(cls, **kwargs: Any) -> Link:
        """
        Tries to figure out what link to use from `kwargs`. E.g.
        if "host" is supplied, then link=Links.SOCKET can be inferred.
        """
        if kwargs.get("host"):
            return Link.SOCKET

        return Link.UART

    @classmethod
    def _try_infer_protocol_from_kwargs(cls, **_: Any) -> Optional[Protocol]:
        """
        Tries to figure out what protocols to use from `kwargs`.
        """

        # Ready for extension
        return None

    @classmethod
    def _try_get_socket_client(
        cls, host: str, protocol: Optional[ProtocolArg], **kwargs: Any
    ) -> BaseClient:
        """
        Tries to create a socket client given the passed `protocol` and `kwargs`.

        :raises: ValueError if it fails somewhere along the way.
        """

        if protocol == Protocol.EXPLORATION or protocol == Protocol.STREAMING or protocol is None:
            # SocketClient auto-detects the protocol.
            return SocketClient(host, **kwargs)
        else:
            raise ValueError(
                f"Could not determine client type. link=socket, protocol={protocol}, host={host}"
            )

    @classmethod
    def _try_get_serial_client(
        cls, port: str, protocol: Optional[ProtocolArg], **kwargs: Any
    ) -> BaseClient:
        """
        Tries to create a serial/uart client given `protocol`, `port` and `kwargs`.

        :raises: ValueError if it fails somewhere along the way.
        """
        if protocol == Protocol.MODULE:
            return UARTClient(port, **kwargs)

        if protocol == Protocol.EXPLORATION:
            return SocketClient(host=port, serial_link=True, **kwargs)

        raise ValueError(
            f"Could not determine client type. link=serial, protocol={protocol}, port={port}"
        )

    @classmethod
    def _handle_passed_host(cls, host: Optional[str]) -> str:
        """
        :raises: ValueError if `host` is None.
        """
        if host is None:
            raise ValueError(
                "Clients that communicate over sockets need a host. Please specify a host with"
                + '"host=<your host>" to "Client"'
            )

        return host

    @classmethod
    def _handle_passed_serial_port(cls, serial_port: Optional[str]) -> str:
        """
        If `serial_port` is None, tries to get a port from auto-detection.

        :raises: ValueError if `serial_port` is None and the number of auto-detected ports != 1.
        """
        if serial_port is None:
            return cls._get_single_acconeer_module_port(et.utils.get_tagged_serial_ports())

        return serial_port

    @classmethod
    def _get_single_acconeer_module_port(cls, tagged_ports: List[Tuple[str, str]]):
        """
        Auto-detects Acconeer modules for use in this factory.

        :raises: ValueError if the number of auto-detected modules is not exactly 1.
        """
        acconeer_module_ports = [port for port, tag in tagged_ports if tag is not None]

        if len(acconeer_module_ports) > 1:
            raise ValueError(
                "Multiple Acconeer modules are connected. Please specify a single port with"
                + '"serial_port=<your serial port>" to "Client"'
            )

        if len(acconeer_module_ports) == 0:
            raise ValueError("Could not auto-detect any Acconeer modules.")

        return acconeer_module_ports[0]



import numpy as np

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et

from .constants import HISTORY_LENGTH_S


INIT_Y_RANGE = 500000


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.processing_config = processing_config

    def setup(self, win):
        win.setWindowTitle("Acconeer button press sparse example")

        dashed_pen = pg.mkPen("k", width=2.5, style=QtCore.Qt.DashLine)

        self.calibrated_text_vis_age = 0

        self.signal_hist_plot = win.addPlot(title="Signal history")
        self.signal_hist_plot.setMenuEnabled(False)
        self.signal_hist_plot.setMouseEnabled(x=False, y=False)
        self.signal_hist_plot.addLegend()
        self.signal_hist_plot.hideButtons()
        self.signal_hist_plot.showGrid(x=True, y=True)
        self.signal_hist_plot.setLabel("bottom", "Time (s)")
        self.signal_hist_plot.setXRange(-HISTORY_LENGTH_S, 0)
        self.signal_hist_plot.setYRange(0, 2**15)

        self.signal_hist_curve = self.signal_hist_plot.plot(
            pen=et.utils.pg_pen_cycler(0),
            name="Signal",
        )
        self.average_hist_curve = self.signal_hist_plot.plot(
            pen=et.utils.pg_pen_cycler(1),
            name="Low pass filtered signal",
        )

        win.nextRow()

        self.proc_plot = win.addPlot(title="Processing history")
        self.proc_plot.setMenuEnabled(False)
        self.proc_plot.setMouseEnabled(x=False, y=False)
        self.proc_plot.hideButtons()
        self.proc_plot.showGrid(x=True, y=True)
        self.proc_plot.setLabel("bottom", "Trigger variable values")
        self.proc_plot.setXRange(-HISTORY_LENGTH_S, 0)
        self.proc_plot.setYRange(0, INIT_Y_RANGE)

        self.trig_hist_curve = self.proc_plot.plot(
            pen=et.utils.pg_pen_cycler(0),
            name="Trig value",
        )

        self.cool_down_hist_curve = self.proc_plot.plot(
            pen=et.utils.pg_pen_cycler(1),
            name="Cool down value",
        )

        self.detection_dots = self.proc_plot.plot(
            pen=None,
            symbol="o",
            symbolSize=20,
            symbolBrush=et.utils.color_cycler(2),
            name="Detections",
        )

        self.limit_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.proc_plot.addItem(self.limit_line)

        self.smooth_limits = et.utils.SmoothLimits(
            self.sensor_config.update_rate, hysteresis=0.1, tau_decay=3, tau_grow=1
        )

        self.smooth_max_proc = et.utils.SmoothMax(
            self.sensor_config.update_rate,
            hysteresis=0.6,
            tau_decay=3,
        )

        self.calibration_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:16pt;">'
            "{}</span></div>"
        )
        calibration_html = self.calibration_html_format.format("Calibrated!")
        self.calibration_text_item = pg.TextItem(
            html=calibration_html,
            fill=pg.mkColor(255, 140, 0),
            anchor=(0.5, 0),
        )

        self.calibration_text_item.setPos(-HISTORY_LENGTH_S / 2, 0.95 * INIT_Y_RANGE)
        self.proc_plot.addItem(self.calibration_text_item)
        self.calibration_text_item.hide()

    def update(self, data):

        signal_hist_ys = data["signal_history"]
        average_hist_ys = data["average_history"]
        cool_down_hist_ys = data["cool_down_history"]
        trig_hist_ys = data["trig_history"]
        t_detections = data["detection_history"]
        calibrated = data["calibrated"]
        hist_xs = np.linspace(-HISTORY_LENGTH_S, 0, len(signal_hist_ys))

        self.signal_hist_curve.setData(hist_xs, signal_hist_ys)
        self.average_hist_curve.setData(hist_xs, average_hist_ys)

        self.trig_hist_curve.setData(hist_xs, trig_hist_ys)
        self.cool_down_hist_curve.setData(hist_xs, cool_down_hist_ys)

        self.detection_dots.setData(t_detections, data["threshold"] * np.ones(len(t_detections)))
        self.limit_line.setPos(data["threshold"])

        limits = self.smooth_limits.update(signal_hist_ys)

        m = 2 * data["threshold"]
        ymax = self.smooth_max_proc.update(m)

        self.signal_hist_plot.setYRange(limits[0], limits[1])
        self.proc_plot.setYRange(0, ymax)

        self.calibration_text_item.setPos(-HISTORY_LENGTH_S / 2, 0.95 * ymax)

        if calibrated:
            self.calibrated_text_vis_age = len(data["signal_history"]) / 10

        if self.calibrated_text_vis_age > 0:
            self.calibration_text_item.setVisible(True)
            self.calibrated_text_vis_age -= 1
        else:
            self.calibration_text_item.setVisible(False)



import argparse
import csv
import json
import os

import numpy as np

import acconeer.exptool as et


DESCRIPTION = """This is a command line utility that lets you convert
.h5/.npz files to .csv-files for use as is or in
e.g. Microsoft Excel.

example usage:
  python3 convert_to_csv.py -v ~/my_data_file.h5 ~/my_output_file.csv
"""


def format_cell_value(v):
    if np.imag(v):
        return f"{np.real(v):0}{np.imag(v):+}j"
    else:
        return str(v)


def parse_config_dump(cfg_str):
    context = {"null": None, "true": True, "false": False}
    return eval(cfg_str, context)


def record_to_csv(
    record: et.a111.recording.Record,
    sensor_index: int = 0,
    sweep_as_column: bool = False,
    add_sweep_metadata: bool = False,
) -> np.ndarray:
    print(record.sensor_config_dump)
    config_dump = parse_config_dump(record.sensor_config_dump)
    print(config_dump)

    print("\nConfiguration")
    for k, v in config_dump.items():
        print(f"{k:30} {v} ")

    num_sensors = record.data.shape[1]
    if sensor_index >= num_sensors:
        raise ValueError(
            f"Invalid sensor index specified (index={sensor_index}). "
            f"Valid indices for this input file is one of {list(range(num_sensors))}"
        )

    # actual translation
    data = record.data[:, sensor_index, :]
    dest_rows = []

    depths = et.a111.get_range_depths(record.sensor_config, record.session_info)

    if record.mode == et.a111.Mode.SPARSE and add_sweep_metadata:
        depths = np.tile(depths, record.sensor_config.sweeps_per_frame)
        sweep_number = np.floor(
            np.linspace(0, record.sensor_config.sweeps_per_frame, num=len(depths), endpoint=False)
        )
        dest_rows.append(sweep_number.astype(int))

    if add_sweep_metadata:
        dest_rows.append(np.round(depths, decimals=6))

    for x in data:
        row = np.ndarray.flatten(x)
        dest_rows.append([format_cell_value(v) for v in row])

    dest = np.array(dest_rows)
    if sweep_as_column:
        dest = dest.T
    return dest


def main():
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _add_arguments(parser)
    args = parser.parse_args()

    input_file = args.input_file
    output_file = args.output_file
    sensor_index = args.index
    force = args.force
    verbose = args.verbose
    delimiter = args.delimiter
    sweep_as_column = args.sweep_as_column
    add_sweep_metadata = args.add_sweep_metadata

    # Convert to real delimiter given to csv module
    if delimiter == "c":
        delimiter = ","
    elif delimiter == "t":
        delimiter = "\t"

    _check_files(input_file, output_file, force)

    print(f'Reading from "{input_file}" ... \n')
    record = et.a111.recording.load(input_file)

    if verbose:
        print("=== Session info " + "=" * 43)
        print_record(record)
        print("=" * 60)
        print()

    try:
        csv_table = record_to_csv(
            record,
            sensor_index=sensor_index,
            sweep_as_column=sweep_as_column,
            add_sweep_metadata=add_sweep_metadata,
        )
        print(f"Writing data with dimensions {csv_table.shape} to {output_file} ...")

        with open(output_file, "w") as f:
            writer = csv.writer(f, delimiter=delimiter)
            for row in csv_table:
                writer.writerow(row)

        print("Success!")
    except ValueError as ve:
        print(ve)
        exit(1)


def _check_files(input_file, output_file, force):
    if not os.path.exists(input_file):
        print(f'The input file ("{input_file}") can not be found.')
        exit(1)

    if os.path.exists(output_file) and not force:
        print(f'The output file ("{output_file}") already exists.')
        print('If you know what you are doing; overwrite it with "-f".')
        exit(1)


def _add_arguments(parser):
    parser.add_argument(
        "-d",
        "--delimiter",
        choices=["c", "t"],
        dest="delimiter",
        default="c",
        help="Delimiter for the output data. Default is comma. 't' is for tab, 'c' for comma",
    )
    parser.add_argument(
        "input_file",
        help='The input file with file endings ".h5" or ".npz".',
    )
    parser.add_argument(
        "output_file",
        help="The output file to which csv-data will be written.",
    )
    parser.add_argument(
        "--index",
        metavar="index",
        dest="index",
        type=int,
        default=0,
        help="The sensor index. Gets data from a specific sensor when multiple sensors are "
        "used (default=0).",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help='Forcefully overwrite "output_file" if it already exists.',
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help='Prints meta data from "input_file".',
    )
    parser.add_argument(
        "--sweep-as-column",
        action="store_true",
        default=False,
        help="Stores sweeps as columns instead of rows.\n"
        "The default is to store sweeps as rows.",
    )
    parser.add_argument(
        "-m",
        "--add_sweep_metadata",
        action="store_true",
        default=False,
        help="Adds depth and sweep number info to the csv file",
    )


def print_record(record):
    print("Mode:", record.mode.name.lower())
    print()
    print(record.sensor_config)
    print()
    print("Session info")

    for k, v in record.session_info.items():
        print("  {:.<35} {}".format(k + " ", v))

    print()
    print("Data shape:", record.data.shape)
    print("Data dtype:", record.data.dtype)
    print()
    print("Last data info (first sensor):")

    for k, v in record.data_info[-1][0].items():
        print("  {:.<35} {}".format(k + " ", v))

    ts = record.sample_times
    if ts is not None and ts.size >= 2:
        print()
        mean_dt = (ts[-1] - ts[0]) / (ts.size - 1)
        mean_f = 1 / mean_dt
        print("Mean sample rate (client side): {:.2f} Hz".format(mean_f))

    print("\n")

    print("Module (processing) key:", record.module_key)

    if record.processing_config_dump is None:
        print("No processing config dump")
    else:
        print("Processing config dump")
        for k, v in json.loads(record.processing_config_dump).items():
            print("  {:.<35} {}".format(k + " ", v))

    print("\n")

    m = {
        "RSS version": record.rss_version,
        "acconeer.exptool library version": record.lib_version,
        "Timestamp": record.timestamp,
    }

    for k, v in m.items():
        print("{:.<37} {}".format(k + " ", v))

    if record.note:
        print()
        print("Note: " + str(record.note))


if __name__ == "__main__":
    main()



import numpy as np

import acconeer.exptool as et


def get_sensor_config():
    config = et.a111.SparseServiceConfig()
    config.profile = et.a111.SparseServiceConfig.Profile.PROFILE_3
    config.sampling_mode = et.a111.SparseServiceConfig.SamplingMode.A
    config.range_interval = [0.24, 0.48]
    config.sweeps_per_frame = 64
    config.sweep_rate = 3e3
    config.hw_accelerated_average_samples = 60
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1

    show_data_plot = et.configbase.BoolParameter(
        label="Show data",
        default_value=True,
        updateable=True,
        order=0,
    )

    show_speed_plot = et.configbase.BoolParameter(
        label="Show speed on FFT y-axis",
        default_value=False,
        updateable=True,
        order=10,
    )


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        pass

    def process(self, data, data_info):
        frame = data

        zero_mean_frame = frame - frame.mean(axis=0, keepdims=True)
        fft = np.fft.rfft(zero_mean_frame.T * np.hanning(frame.shape[0]), axis=1)
        abs_fft = np.abs(fft)

        return {
            "frame": frame,
            "abs_fft": abs_fft,
        }



import logging
import multiprocessing as mp
import platform
import queue
import signal
import socket
import traceback
from abc import ABCMeta, abstractmethod
from time import sleep, time

import serial

from acconeer.exptool.a111._clients.base import ClientError


log = logging.getLogger(__name__)


class LinkError(ClientError):
    pass


class BaseLink(metaclass=ABCMeta):
    DEFAULT_TIMEOUT = 2

    @abstractmethod
    def __init__(self):
        self._timeout = self.DEFAULT_TIMEOUT

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def recv(self, num_bytes):
        pass

    @abstractmethod
    def recv_until(self, bs):
        pass

    @abstractmethod
    def send(self, data):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, val):
        self._timeout = val
        self._update_timeout()

    def _update_timeout(self):
        pass


class SocketLink(BaseLink):
    _CHUNK_SIZE = 4096
    _PORT = 6110

    def __init__(self, host=None):
        super().__init__()
        self._host = host
        self._sock = None
        self._buf = None

    def _update_timeout(self):
        if self._sock is not None:
            self._sock.settimeout(self._timeout)

    def connect(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._update_timeout()

        try:
            self._sock.connect((self._host, self._PORT))
        except OSError as e:
            self._sock = None
            raise LinkError("failed to connect") from e

        self._buf = bytearray()

    def recv(self, num_bytes):
        while len(self._buf) < num_bytes:
            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs):
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data

    def send(self, data):
        self._sock.sendall(data)

    def disconnect(self):
        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()
        self._sock = None
        self._buf = None


class BaseSerialLink(BaseLink):
    def __init__(self, baudrate=115200):
        super().__init__()
        self._timeout = self.DEFAULT_TIMEOUT
        self._baudrate = baudrate


class SerialLink(BaseSerialLink):
    def __init__(self, port=None, flowcontrol=False):
        super().__init__()
        self._port = port
        self._ser = None
        self._flowcontrol = flowcontrol

    def _update_timeout(self):
        if self._ser is not None:
            self._ser.timeout = self._timeout

    def send_break(self):
        self._ser.send_break()
        sleep(0.5)
        self._ser.flushInput()

    def connect(self):
        self._ser = serial.Serial()
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._update_timeout()
        self._ser.open()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

    def recv(self, num_bytes):
        data = bytearray(self._ser.read(num_bytes))

        if not len(data) == num_bytes:
            raise LinkError("recv timeout")

        return data

    def recv_until(self, bs):
        data = bytearray(self._ser.read_until(bs))

        if bs not in data:
            raise LinkError("recv timeout")

        return data

    def send(self, data):
        self._ser.write(data)

    def disconnect(self):
        self._ser.close()
        self._ser = None

    @property
    def baudrate(self):
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate):
        self._baudrate = new_baudrate

        if self._ser is not None and self._ser.is_open:
            self._ser.baudrate = new_baudrate


class ExploreSerialLink(SerialLink):
    _SERIAL_READ_PACKET_SIZE = 65536
    _SERIAL_PACKET_TIMEOUT = 0.01

    def __init__(self, port, flowcontrol=True):
        super().__init__(port, flowcontrol)
        self._buf = None

    def _update_timeout(self):
        pass

    def connect(self):
        self._ser = serial.Serial(timeout=self._SERIAL_PACKET_TIMEOUT)
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._ser.open()
        self._buf = bytearray()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

        self.send_break()

    def recv(self, num_bytes):
        t0 = time()
        while len(self._buf) < num_bytes:
            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs):
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data


class SerialProcessLink(BaseSerialLink):
    def __init__(self, port=None):
        super().__init__()
        self._port = port
        self._process = None

    def connect(self):
        self._recv_queue = mp.Queue()
        self._send_queue = mp.Queue()
        self._flow_event = mp.Event()
        self._error_event = mp.Event()

        args = (
            self._port,
            self._baudrate,
            self._recv_queue,
            self._send_queue,
            self._flow_event,
            self._error_event,
        )

        self._process = mp.Process(
            target=serial_process_program,
            args=args,
            daemon=True,
        )

        self._process.start()

        flow_event_was_set = self._flow_event.wait(self._timeout)

        if flow_event_was_set:
            log.debug("connect - flow event was set")
        else:
            log.debug("connect - flow event was not set (timeout)")
            self.disconnect()
            raise LinkError("failed to connect, timeout")

        if self._error_event.is_set():
            log.debug("connect - error event was set")
            self.disconnect()
            raise LinkError("failed to connect, see traceback from serial process")

        log.debug("connect - successful")

        self._buf = bytearray()

    def recv(self, num_bytes):
        self.__empty_queue_into_buf()

        t0 = time()
        while len(self._buf) < num_bytes:
            self.__get_into_buf()

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs):
        self.__empty_queue_into_buf()

        n = len(bs)
        si = 0
        t0 = time()
        while True:
            buf_size = len(self._buf)
            if buf_size >= n:
                try:
                    i = self._buf.index(bs, si)
                except ValueError:
                    pass
                else:
                    break

                si = buf_size - n + 1

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            self.__get_into_buf()

        data = self._buf[: i + n]
        self._buf = self._buf[i + n :]
        return data

    def send(self, data):
        self._send_queue.put(data)

    def disconnect(self):
        if self._process.exitcode is None:
            self._flow_event.clear()
            self._process.join(1)

        if self._process.exitcode is None:
            log.info("forcing disconnect...")
            self._process.terminate()
            self._process.join(1)
            log.info("forced disconnect successful")

        if self._process.exitcode is None:
            raise LinkError("failed to disconnect")

    def __empty_queue_into_buf(self):
        while True:
            try:
                data = self._recv_queue.get_nowait()
            except queue.Empty:
                break

            self._buf.extend(data)

    def __get_into_buf(self):
        try:
            data = self._recv_queue.get(timeout=self._timeout)
        except queue.Empty:
            raise LinkError("recv timeout")
        except InterruptedError:
            return  # fixes interrupt issue on Windows

        self._buf.extend(data)

    @property
    def baudrate(self):
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate):
        self._baudrate = new_baudrate

        if self._process is not None and self._process.exitcode is None:
            log.debug("Changing baudrate to {}".format(new_baudrate))
            self._send_queue.put(("baudrate", new_baudrate))


def serial_process_program(port, baud, recv_q, send_q, flow_event, error_event):
    log.debug("serial communication process started")
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    try:
        _serial_process_program(port, baud, recv_q, send_q, flow_event, error_event)
    except Exception:
        error_event.set()
        flow_event.set()
        sleep(0.1)  # give the main process some time to print log messages
        print("Exception raised in serial process:\n")
        traceback.print_exc()
        print("\n\n")

    recv_q.close()
    send_q.close()


def _serial_process_program(port, baud, recv_q, send_q, flow_event, error_event):
    ser = serial.Serial(port=port, baudrate=baud, timeout=0, exclusive=True)
    flow_event.set()
    while flow_event.is_set():
        received = bytearray()
        while True:
            data = bytearray(ser.read(4096))
            if len(data) == 0:
                break
            received.extend(data)
        if len(received) > 0:
            recv_q.put(received)

        sent = False
        while True:
            try:
                x = send_q.get_nowait()
            except queue.Empty:
                break

            if isinstance(x, tuple):
                _, val = x  # assume its a baudrate change
                ser.baudrate = val
            else:
                ser.write(x)
                sent = True

        if not sent:
            sleep(0.0025)

    ser.close()



import enum
import json

import acconeer.exptool._structs.configbase as cb
from acconeer.exptool import utils
from acconeer.exptool.a111._modes import Mode, get_mode


class ConfigEnum(enum.Enum):
    @property
    def label(self):
        return self.value[0]

    @property
    def json_value(self):
        return self.value[1]


class ModeParameter(cb.ConstantParameter):
    def __init__(self, **kwargs):
        assert isinstance(kwargs["value"], Mode)
        kwargs.setdefault("does_dump", True)
        super().__init__(**kwargs)

    def dump(self, obj):
        return self.__get__(obj).name

    def load(self, obj, value):
        assert get_mode(value) == self.value


class BaseSessionConfig(cb.SensorConfig):
    sensor = cb.SensorParameter(
        label="Sensor(s)",
        default_value=[1],
        order=0,
        help=r"""
            The sensor(s) to be configured.
        """,
    )


class BaseServiceConfig(BaseSessionConfig):
    class RepetitionMode(ConfigEnum):
        HOST_DRIVEN = ("Host driven", "on_demand")
        SENSOR_DRIVEN = ("Sensor driven", "streaming")

    class Profile(ConfigEnum):
        PROFILE_1 = ("1 (max resolution)", 1, 0.10)
        PROFILE_2 = ("2", 2, 0.12)
        PROFILE_3 = ("3", 3, 0.18)
        PROFILE_4 = ("4", 4, 0.36)
        PROFILE_5 = ("5 (max SNR)", 5, 0.60)

        @property
        def approx_direct_leakage_length(self):
            return self.value[2]

    class PowerSaveMode(ConfigEnum):
        ACTIVE = ("Active", "active")
        READY = ("Ready", "ready")
        SLEEP = ("Sleep", "sleep")
        HIBERNATE = ("Hibernate", "hibernate")
        OFF = ("Off", "off")

    range_interval = cb.FloatRangeParameter(
        label="Range interval",
        unit="m",
        default_value=[0.18, 0.78],
        limits=(-0.7, 7.0),
        order=10,
        help=r"""
            The measured depth range. The start and end values will be rounded to the closest
            measurement point available.
        """,
    )

    range_start = cb.get_virtual_parameter_class(cb.FloatParameter)(
        label="Range start",
        get_fun=lambda conf: conf.range_interval[0],
        visible=False,
    )

    range_length = cb.get_virtual_parameter_class(cb.FloatParameter)(
        label="Range length",
        get_fun=lambda conf: conf.range_interval[1] - conf.range_interval[0],
        visible=False,
    )

    range_end = cb.get_virtual_parameter_class(cb.FloatParameter)(
        label="Range end",
        get_fun=lambda conf: conf.range_interval[1],
        visible=False,
    )

    repetition_mode = cb.EnumParameter(
        label="Repetition mode",
        enum=RepetitionMode,
        default_value=RepetitionMode.HOST_DRIVEN,
        order=1010,
        category=cb.Category.ADVANCED,
        help=r"""
            The RSS supports two different repetition modes. They determine how and when data
            acquisition occurs. They are:

            * **On demand / host driven**: The sensor produces data when requested by the
              application. Hence, the application is responsible for timing the data acquisition.
              This is the default mode, and may be used with all power save modes.

            * **Streaming / sensor driven**: The sensor produces data at a fixed rate, given by a
              configurable accurate hardware timer. This mode is recommended if exact timing
              between updates is required.

            The Exploration Tool is capable of setting the update rate also in *on demand (host
            driven)* mode. Thus, the difference between the modes becomes subtle. This is why *on
            demand* and *streaming* are called *host driven* and *sensor driven* respectively in
            Exploration Tool.
        """,
    )

    update_rate = cb.FloatParameter(
        label="Update rate",
        unit="Hz",
        default_value=None,
        limits=(0.1, None),
        decimals=1,
        optional=True,
        optional_label="Limit",
        optional_default_set_value=50.0,
        order=30,
        help=r"""
            The rate :math:`f_f` at which the sensor sends frames to the host MCU.

            .. attention::

               Setting the update rate too high might result in missed data frames.

            In sparse, the maximum possible update rate depends on the *sweeps per frame*
            :math:`N_s` and *sweep rate* :math:`f_s`:

            .. math::

               \frac{1}{f_f} > N_s \cdot \frac{1}{f_s} + \text{overhead*}

            \* *The overhead largely depends on data frame size and data transfer speeds.*
        """,
    )

    gain = cb.FloatParameter(
        label="Gain",
        default_value=0.5,
        limits=(0.0, 1.0),
        decimals=2,
        order=1040,
        category=cb.Category.ADVANCED,
        help=r"""
            The receiver gain used in the sensor. If the gain is too low, objects may not be
            visible, or it may result in poor signal quality due to quantization errors. If the
            gain is too high, strong reflections may result in saturated data. We recommend not
            setting the gain higher than necessary due to signal quality reasons.

            Must be between 0 and 1 inclusive, where 1 is the highest possible gain.

            .. note::
               When Sensor normalization is active, the change in the data due to changing gain is
               removed after normalization. Therefore, the data might seen unaffected by changes
               in the gain, except very high (receiver saturation) or very low (quantization
               error) gain.

               Sensor normalization is not available for the Sparse service, but is enabled by
               default for the other services - Envelope, IQ, and Power Bins.
        """,
    )

    hw_accelerated_average_samples = cb.IntParameter(
        label="HW accel. average samples",
        default_value=10,
        limits=(1, 63),
        order=1030,
        category=cb.Category.ADVANCED,
        help=r"""
            Number of samples taken to obtain a single point in the data. These are averaged
            directly in the sensor hardware - no extra computations are done in the MCU.

            The time needed to measure a sweep is roughly proportional to the HWAAS. Hence, if
            there's a need to obtain a higher sweep rate, HWAAS could be decreased. Note that
            HWAAS does not affect the amount of data transmitted from the sensor over SPI.

            Must be at least 1 and not greater than 63.
        """,
    )

    maximize_signal_attenuation = cb.BoolParameter(
        label="Max signal attenuation",
        default_value=False,
        order=2000,
        category=cb.Category.ADVANCED,
        help=r"""
            When measuring in the direct leakage (around 0m), this setting can be enabled to
            minimize saturation in the receiver. We do not recommend using this setting under
            normal operation.
        """,
    )

    profile = cb.EnumParameter(
        label="Profile",
        enum=Profile,
        default_value=Profile.PROFILE_2,
        order=20,
        help=r"""
            The main configuration of all the services are the profiles, numbered 1 to 5. The
            difference between the profiles is the length of the radar pulse and the way the
            incoming pulse is sampled. Profiles with low numbers use short pulses while the higher
            profiles use longer pulses.

            Profile 1 is recommended for:

            - measuring strong reflectors, to avoid saturation of the received signal
            - close range operation (<20 cm), due to the reduced direct leakage

            Profile 2 and 3 are recommended for:

            - operation at intermediate distances, (20 cm to 1 m)
            - where a balance between SNR and depth resolution is acceptable

            Profile 4 and 5 are recommended for:

            - for Sparse service only
            - operation at large distances (>1 m)
            - motion or presence detection, where an optimal SNR ratio is preferred over a high
              resolution distance measurement

            The previous profile Maximize Depth Resolution and Maximize SNR are now profile 1 and
            2. The previous Direct Leakage Profile is obtained by the use of the Maximize Signal
            Attenuation parameter.
        """,
    )

    downsampling_factor = cb.IntParameter(
        label="Downsampling factor",
        default_value=1,
        limits=(1, None),
        order=1020,
        category=cb.Category.ADVANCED,
        help=r"""
            The range downsampling by an integer factor. A factor of 1 means no downsampling, thus
            sampling with the smallest possible depth interval. A factor 2 samples every other
            point, and so on. In Envelope and IQ, the finest interval is ~0.5 mm. In Power Bins,
            it is the same but then further downsampled in post-processing.
            In sparse, it is ~6 cm.

            The downsampling is performed by skipping measurements in the sensor, and therefore
            gives lower memory usage, lower power consumption, and lower duty cycle.

            In sparse, setting a too large factor might result in gaps in the data where moving
            objects "disappear" between sampling points.

            In Envelope, IQ, and Power Bins, the factor must be 1, 2, or 4.
            In sparse, it must be at least 1.
            Setting a factor greater than 1 might affect the range end point and for IQ and
            Envelope, also the first point.
        """,
    )

    tx_disable = cb.BoolParameter(
        label="Disable TX",
        default_value=False,
        order=3000,
        category=cb.Category.ADVANCED,
        help=r"""
            Disable the radio transmitter. If used to measure noise, we recommended also switching
            off noise level normalization (if applicable).
        """,
    )

    power_save_mode = cb.EnumParameter(
        label="Power save mode",
        enum=PowerSaveMode,
        default_value=PowerSaveMode.ACTIVE,
        order=3100,
        category=cb.Category.ADVANCED,
        help=r"""
            The power save mode configuration sets what state the sensor waits in between
            measurements in an active service. There are five power save modes. The modes
            differentiate in current dissipation and response latency, where the most current
            consuming mode *Active* gives fastest response and the least current consuming mode
            *Off* gives the slowest response. The absolute response time and also maximum update
            rate is determined by several factors besides the power save mode configuration.

            In addition, the host capabilities in terms of SPI communication speed and
            processing speed also impact on the absolute response time. The power consumption of
            the system depends on the actual configuration of the application and it is recommended
            to test both maximum update rate and power consumption when the configuration
            is decided.

            ================== ==================== ==============
            Power save mode    Current consumption  Response time
            ================== ==================== ==============
            Off                Lowest               Longest
            Hibernate          ...                  ...
            Sleep              ...                  ...
            Ready              ...                  ...
            Active             Highest              Shortest
            ================== ==================== ==============

            .. note::
               Hibernation has limited hardware support. It is not supported by the Raspberry Pi
               EVK:s and XM112.
        """,
    )

    asynchronous_measurement = cb.BoolParameter(
        label="Enable asynchronous measurement",
        default_value=True,
        order=3200,
        category=cb.Category.ADVANCED,
        help=r"""
            Enabling asynchronous measurements will result in a faster update rate but introduces a
            risk of interference between sensors.
        """,
    )

    def check(self):
        alerts = []

        if self.repetition_mode == __class__.RepetitionMode.SENSOR_DRIVEN:
            msg = "Must be set when sensor driven"

            if self.update_rate is None:
                alerts.append(cb.Error("update_rate", msg))

            if not self.asynchronous_measurement:
                alerts.append(cb.Error("asynchronous_measurement", msg))

        if self.gain > 0.9:
            alerts.append(cb.Warning("gain", "Too high gain causes degradation"))

        if self.range_start < self.profile.approx_direct_leakage_length:
            alerts.append(cb.Info("range_interval", "Direct leakage might be seen"))

        if self.power_save_mode == __class__.PowerSaveMode.HIBERNATE:
            alerts.append(cb.Warning("power_save_mode", "Limited hardware support"))

            if self.repetition_mode == __class__.RepetitionMode.SENSOR_DRIVEN:
                alerts.append(cb.Error("power_save_mode", "Unavailable when sensor driven"))

        psms = [__class__.PowerSaveMode.HIBERNATE, __class__.PowerSaveMode.OFF]
        if self.power_save_mode in psms:
            if self.asynchronous_measurement:
                msg = "PSM hibernate/off is always synchronous"
                alerts.append(cb.Info("asynchronous_measurement", msg))

        return alerts


class BaseDenseServiceConfig(BaseServiceConfig):
    noise_level_normalization = cb.BoolParameter(
        label="Noise level normalization",
        default_value=True,
        order=2010,
        category=cb.Category.ADVANCED,
        help=r"""
            With the SW version 2 release, a sensor signal normalization functionality is activated
            by default for the Power Bins, Envelope, and IQ Service. This results in a more
            constant signal for different temperatures and sensors. The radar sweeps are normalized
            to have similar amplitude independent of sensor gain and hardware averaging, resulting
            in only minor visible effect in the sweeps when adjusting these parameters.

            We recommend this setting especially for applications, where absolute radar amplitudes
            are important, such as when comparing to a previously recorded signal or to a fixed
            threshold.

            More technically, the functionality is implemented to collect data when starting the
            service, but not transmitting pulses. This data is then used to determine the current
            sensitivity of receiving part of the radar by estimating the power level of the noise,
            which then is used to normalize the collected sweeps. In the most low-power systems,
            where a service is created to collect just a single short sweep before turning off, the
            sensor normalization can add a non-negligible part to the power consumption.

            Please note, that due to the nature of Sparse data, the Sparse service does not support
            noise level normalization. Instead, normalization during processing is recommended,
            such as done in the Presence detector.

        """,
    )

    def check(self):
        alerts = super().check()

        if self.range_length < 1e-6:
            alerts.append(cb.Warning("range_interval", "Will only return a single point"))

        if self.downsampling_factor not in [1, 2, 4]:
            alerts.append(cb.Error("downsampling_factor", "Must be 1, 2, or 4"))

        if self.repetition_mode == __class__.RepetitionMode.SENSOR_DRIVEN:
            chunks = self.range_length / 0.06
            chunks += 2 if self.mode == Mode.IQ else 0
            points_per_chunk = 124 / self.downsampling_factor
            if points_per_chunk * chunks > 2048:
                alerts.append(cb.Error("range_interval", "Too large for buffer"))

        if self.power_save_mode == __class__.PowerSaveMode.HIBERNATE:
            alerts.append(cb.Error("power_save_mode", "Not supported for this service"))

        return alerts


class _MURCapable(cb.SensorConfig):
    class MUR(ConfigEnum):
        MUR_6 = ("6 (MMD: 7.0m)", 6, 7.0)
        MUR_9 = ("9 (MMD: 12.7m)", 9, 12.7)

        @property
        def mmd(self):
            return self.value[2]

    mur = cb.EnumParameter(
        label="Max. unambiguous range",
        enum=MUR,
        default_value=MUR.MUR_6,
        order=15,
        help=r"""
            Sets the *maximum unambiguous range* (MUR), which in turn sets the *maximum measurable
            distance* (MMD).

            The MMD is the maximum value for the range end, i.e., the range start + length. The MMD
            is smaller than the MUR due to hardware limitations.

            The MUR is the maximum distance at which an object can be located to guarantee that its
            reflection corresponds to the most recent transmitted pulse. Objects farther away than
            the MUR may fold into the measured range. For example, with a MUR of 10 m, an object at
            12 m could become visible at 2 m.

            A higher setting gives a larger MUR/MMD, but comes at a cost of increasing the
            measurement time for a sweep. The measurement time is approximately proportional to the
            MUR.

            This setting changes the *pulse repetition frequency* (PRF) of the radar system. The
            relation between PRF and MUR is

            .. math::

               \text{MUR} = c / (2 * \text{PRF})

            where *c* is the speed of light.

            =======  ======  ======  ========
            Setting     MUR     MMD       PRF
            =======  ======  ======  ========
                  6  11.5 m   7.0 m  13.0 MHz
                  9  17.3 m  12.7 m   8.7 MHz
            =======  ======  ======  ========

            This is an experimental feature.
        """,
    )

    range_interval = cb.FloatRangeParameter(
        label="Range interval",
        unit="m",
        default_value=[0.18, 0.78],
        limits=(-0.7, 12.7),
        order=10,
        help=r"""
            The measured depth range. The start and end values will be rounded to the closest
            measurement point available.

            The the sweep range is limited to 7.0 m for the default "maximum unambiguous range"
            setting. To change this limitation, increase "maximum unambiguous range".
        """,
    )

    def check(self):
        alerts = super().check()
        if self.mur.mmd < self.range_end:
            alerts.append(
                cb.Error(
                    "mur",
                    "Too low for the given range",
                )
            )
            alerts.append(
                cb.Error(
                    "range_interval",
                    f"MUR limits the range to {self.mur.mmd:.1f} m",
                )
            )
        return alerts


class PowerBinServiceConfig(_MURCapable, BaseDenseServiceConfig):
    _MIN_BIN_SIZE = 0.016

    mode = ModeParameter(
        label="Mode",
        value=Mode.POWER_BINS,
    )

    bin_count = cb.IntParameter(
        label="Bin count",
        default_value=5,
        limits=(1, None),
        order=5,
        help=r"""
            The number of bins to be used for creating the amplitude over distance histogram.
        """,
    )

    def check(self):
        alerts = super().check()

        if self.range_length < self._MIN_BIN_SIZE:
            alerts.append(cb.Error("range_interval", "Too short"))
        elif self.range_length / self.bin_count < self._MIN_BIN_SIZE:
            alerts.append(cb.Error("bin_count", "Too high for current range"))

        return alerts


class EnvelopeServiceConfig(_MURCapable, BaseDenseServiceConfig):
    mode = ModeParameter(
        label="Mode",
        value=Mode.ENVELOPE,
    )

    running_average_factor = cb.FloatParameter(
        label="Running avg. factor",
        default_value=0.7,
        limits=(0.0, 1.0),
        order=500,
    )

    def check(self):
        alerts = super().check()
        return alerts


class IQServiceConfig(BaseDenseServiceConfig):
    class SamplingMode(ConfigEnum):
        A = ("A (less correlation)", 0)
        B = ("B (more SNR)", 1)

    mode = ModeParameter(
        label="Mode",
        value=Mode.IQ,
    )

    sampling_mode = cb.EnumParameter(
        label="Sampling mode",
        enum=SamplingMode,
        default_value=SamplingMode.A,
        order=1000,
        category=cb.Category.ADVANCED,
        is_dummy=True,
    )

    depth_lowpass_cutoff_ratio = cb.FloatParameter(
        label="Depth LPF cutoff ratio",
        default_value=None,
        limits=(0.0, 0.5),
        decimals=6,
        optional=True,
        optional_default_set_value=0.5,
        optional_label="Override",
        order=2100,
        category=cb.Category.ADVANCED,
        help=r"""
            Depth domain lowpass filter cutoff frequency ratio

            The cutoff for the depth domain lowpass filter is specified as the ratio between the
            spatial frequency cutoff and the sample frequency. A ratio of zero ratio will configure
            the smoothest possible filter. A ratio of 0.5 (the Nyquist frequency) turns the filter
            off.

            If unset, i.e., if not overridden, the ratio will be chosen automatically. The used
            ratio is returned in the session information (metadata) upon session setup (create).
        """,
    )

    _depth_lowpass_cutoff_ratio_value = cb.get_virtual_parameter_class(cb.FloatParameter)(
        label="Depth LPF cutoff ratio value",
        get_fun=lambda conf: utils.optional_or_else(conf.depth_lowpass_cutoff_ratio, 0.0),
        visible=False,
    )

    _depth_lowpass_cutoff_ratio_override = cb.get_virtual_parameter_class(cb.BoolParameter)(
        label="Depth LPF cutoff ratio override",
        get_fun=lambda conf: conf.depth_lowpass_cutoff_ratio is not None,
        visible=False,
    )

    def check(self):
        alerts = super().check()

        if self.range_start < (0.06 - 1e-6):
            alerts.append(cb.Error("range_interval", "Start must be >= 0.06 m"))

        if self.sampling_mode == __class__.SamplingMode.B:
            alerts.append(cb.Error("sampling_mode", "IQ sampling mode B is removed"))

        return alerts


class SparseServiceConfig(_MURCapable, BaseServiceConfig):
    class SamplingMode(ConfigEnum):
        A = ("A (less correlation)", 0)
        B = ("B (more SNR)", 1)

    mode = ModeParameter(
        label="Mode",
        value=Mode.SPARSE,
    )

    sweeps_per_frame = cb.IntParameter(
        label="Sweeps per frame",
        default_value=16,
        limits=(1, 2048),
        order=50,
        help=r"""
            The number of sweeps per frame :math:`N_s`.

            Must be at least 1, and not greater than 64 when using sampling mode B.
        """,
    )

    sweep_rate = cb.FloatParameter(
        label="Sweep rate",
        unit="Hz",
        default_value=None,
        limits=(1, None),
        decimals=0,
        optional=True,
        optional_default_set_value=3000.0,
        order=40,
        help=r"""
            In Sparse, each frame is a collection of several sweeps over the selected distance
            range (sweeps per frame). The sweep rate :math:`f_s` is the rate at which sweeps are
            performed, i.e. the rate at which each distance point is scanned. If you set the sweep
            rate to 4000 Hz and the sweeps per frame to 32, each Sparse data frame will contain 32
            sweeps over the selected distance range, where the sweeps are measured at a rate of
            4000 Hz.

            The maximum possible sweep rate...

            - Is roughly inversely proportional to the number of depth points measured (affected by
              the **range interval** and **downsampling factor**).
            - Is roughly inversely proportional to **HW accelerated average samples**.
            - Depends on the **sampling mode**. Mode A is roughly :math:`4/3 \approx 130\%` slower
              than mode B with the same configuration.

            To get the maximum possible rate, leave this value unset and look at the :ref:`sweep
            rate <sparse-info-sweep-rate>` in the session info (metadata).

            .. tip::
               If you do not need a specific sweep rate, we recommend leaving it unset.
        """,
    )

    sampling_mode = cb.EnumParameter(
        label="Sampling mode",
        enum=SamplingMode,
        default_value=SamplingMode.B,
        order=1000,
        category=cb.Category.ADVANCED,
        help=r"""
            The sampling mode changes how the hardware accelerated averaging is done.
            This may either increase SNR or reduce correlation.

            *Mode A* is:

            - optimized for maximal independence of the depth points, giving a higher depth
              resolution than mode B.
            - more suitable for applications like gesture recognition, measuring the distance to a
              movement, and speed measurements.

            *Mode B* is:

            - optimized for maximal SNR per unit time spent on measuring. This makes it more energy
              efficient and suitable for cases where small movements are to be detected over long
              ranges.
            - resulting in roughly 3 dB better SNR per unit time than mode A.
        """,
    )

    def check(self):
        alerts = super().check()

        if self.sampling_mode == __class__.SamplingMode.B:
            if self.sweeps_per_frame > 64:
                alerts.append(cb.Error("sweeps_per_frame", "Must be < 64 with sampling mode B"))

        if self.sweep_rate is not None and self.update_rate is not None:
            max_frame_rate = self.sweep_rate / self.sweeps_per_frame

            if self.update_rate > max_frame_rate:
                alerts.append(cb.Error("sweep_rate", "Too low for current update rate"))

        if self.downsampling_factor not in [1, 2, 4, 8]:
            alerts.append(cb.Warning("downsampling_factor", "Must be 1, 2, 4, or 8"))

        # Pessimistic buffer size check
        start_p = int(round(self.range_start / 0.06 - 0.01))
        end_p = int(round(self.range_end / 0.06 + 0.01))
        sweep_size = (end_p - start_p) // self.downsampling_factor + 1
        if sweep_size * self.sweeps_per_frame > 2048:
            alerts.append(cb.Error("range_interval", "Too long for buffer"))

        return alerts


MODE_TO_CONFIG_CLASS_MAP = {
    Mode.POWER_BINS: PowerBinServiceConfig,
    Mode.ENVELOPE: EnvelopeServiceConfig,
    Mode.IQ: IQServiceConfig,
    Mode.SPARSE: SparseServiceConfig,
}


def loads(dump: str, mode=None) -> cb.SensorConfig:
    if mode is None:
        mode = json.loads(dump)["mode"]

    mode = get_mode(mode)
    config = MODE_TO_CONFIG_CLASS_MAP[mode]()
    config._loads(dump)
    return config


load = loads  # for backwards compatibility


def dumps(config: cb.SensorConfig) -> str:
    return config._dumps()



import os
import sys

import acconeer.exptool as et


def main():
    parser = et.a111.ExampleArgumentParser()
    parser.add_argument("-o", "--output-dir", type=str, required=True)
    parser.add_argument("--file-format", type=str, default="h5")
    parser.add_argument("--frames-per-file", type=int, default=10000)
    args = parser.parse_args()
    et.utils.config_logging(args)

    if os.path.exists(args.output_dir):
        print("Directory '{}' already exists, won't overwrite".format(args.output_dir))
        sys.exit(1)

    file_format = args.file_format.lower()
    if file_format == "np":
        file_format = "npz"

    if file_format not in ["h5", "npz"]:
        print("Unknown format '{}'".format(args.file_format))
        sys.exit(1)

    if args.frames_per_file < 10:
        print("Frames per file must be at least 10")
        sys.exit(1)

    client = et.a111.Client(**et.a111.get_client_args(args))

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = args.sensors
    config.update_rate = 30

    session_info = client.start_session(config)

    os.makedirs(args.output_dir)

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    total_num_frames = 0
    while not interrupt_handler.got_signal:
        record_count, num_frames_in_record = divmod(total_num_frames, args.frames_per_file)

        if num_frames_in_record == 0:
            recorder = et.a111.recording.Recorder(sensor_config=config, session_info=session_info)

        data_info, data = client.get_next()
        recorder.sample(data_info, data)

        if num_frames_in_record + 1 == args.frames_per_file:
            record = recorder.close()
            filename = os.path.join(
                args.output_dir, "{:04}.{}".format(record_count + 1, file_format)
            )
            print("Saved", filename)
            et.a111.recording.save(filename, record)

        total_num_frames += 1
        print("Sampled {:>5}".format(total_num_frames), end="\r", flush=True)

    try:
        client.disconnect()
    except Exception:
        pass

    record_count, num_frames_in_record = divmod(total_num_frames, args.frames_per_file)
    if num_frames_in_record > 0:
        record = recorder.close()
        filename = os.path.join(args.output_dir, "{:04}.{}".format(record_count + 1, file_format))
        print("Saved", filename)
        et.a111.recording.save(filename, record)


if __name__ == "__main__":
    main()



import sys
from pathlib import Path
from subprocess import DEVNULL, CalledProcessError, check_call


def main():
    here = Path(__file__).resolve().parent

    try:
        check_call([sys.executable, "-m", "pip", "-V"], stdout=DEVNULL, stderr=DEVNULL)
    except CalledProcessError:
        check_call([sys.executable, here / "get-pip.py", "--no-warn-script-location"])

    install_cmd = [
        sys.executable,
        "-m",
        "pip",
        "install",
        "--no-warn-script-location",
        "--no-input",
        "--upgrade",
    ]

    if Path("testpypi").is_file():
        install_cmd.extend(
            [
                "--index-url",
                "https://test.pypi.org/simple/",
                "--extra-index-url",
                "https://pypi.org/simple/",
            ]
        )

    install_cmd.append("acconeer-exptool[app]")

    check_call(install_cmd)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Optional

import numpy as np

import acconeer.exptool as et
from acconeer.exptool.a111.algo import (
    Calibration,
    CalibrationMapper,
    _AcceptedFileExtensions,
    _Path,
)


@dataclass(frozen=True)
class EnvelopeCalibration(Calibration):
    background: np.ndarray

    def save(self, path: _Path):
        self.validate_path(path)
        np.save(path, self.background)

    @classmethod
    def load(cls, path: _Path) -> Calibration:
        cls.validate_path(path)
        background = np.load(path)
        return EnvelopeCalibration(background)

    @classmethod
    def file_extensions(cls) -> _AcceptedFileExtensions:
        return [("npy", "Numpy data files (*.npy)")]


class EnvelopeCalibrationConfiguration(et.configbase.Config):
    pass


class EnvelopeCalibrationMapper(CalibrationMapper):
    @classmethod
    def get_updated_calibration_from_configuration(
        cls,
        configuration: EnvelopeCalibrationConfiguration,
        calibration: Optional[EnvelopeCalibration],
    ) -> EnvelopeCalibration:
        if calibration is None:
            raise ValueError("Calibration cannot be None in this context.")

        return calibration

    @classmethod
    def update_config_from_calibration(
        cls,
        configuration: EnvelopeCalibrationConfiguration,
        calibration: EnvelopeCalibration,
    ) -> None:
        pass



from acconeer.exptool.a111.algo import ModuleFamily, ModuleInfo

from ._processor import ProcessingConfiguration, Processor, get_sensor_config
from .calibration import (
    ObstacleDetectionCalibration,
    ObstacleDetectionCalibrationConfiguration,
    ObstacleDetectionCalibrationMapper,
)
from .ui import PGUpdater


module_info = ModuleInfo(
    key="iq_obstacle",
    label="Obstacle detection (IQ)",
    pg_updater=PGUpdater,
    processing_config_class=ProcessingConfiguration,
    module_family=ModuleFamily.DETECTOR,
    sensor_config_class=get_sensor_config,
    processor=Processor,
    multi_sensor=[1, 2],
    docs_url="https://docs.acconeer.com/en/latest/processing/obstacle.html",
    calibration_class=ObstacleDetectionCalibration,
    calibration_config_class=ObstacleDetectionCalibrationConfiguration,
    calibration_mapper=ObstacleDetectionCalibrationMapper,
)



import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.config = sensor_config

    def setup(self, win):
        win.resize(800, 600)
        win.setWindowTitle("Acconeer sleep breathing estimation example")

        phi_title = "Breathing motion (detection range: {} m to {} m)".format(
            *self.config.range_interval
        )
        self.phi_plot = win.addPlot(title=phi_title)
        self.phi_plot.setMenuEnabled(False)
        self.phi_plot.setMouseEnabled(x=False, y=False)
        self.phi_plot.hideButtons()
        self.phi_plot.showGrid(x=True, y=True)
        self.phi_plot.setLabel("left", "Amplitude")
        self.phi_plot.setLabel("bottom", "Samples")
        self.phi_plot.addLegend()
        self.filt_phi_curve = self.phi_plot.plot(
            pen=et.utils.pg_pen_cycler(0),
            name="Filtered",
        )
        self.raw_phi_curve = self.phi_plot.plot(
            pen=et.utils.pg_pen_cycler(1),
            name="Raw",
        )

        win.nextRow()

        self.spect_plot = win.addPlot(title="Power spectrum")
        self.spect_plot.setMenuEnabled(False)
        self.spect_plot.setMouseEnabled(x=False, y=False)
        self.spect_plot.hideButtons()
        self.spect_plot.showGrid(x=True, y=True)
        self.spect_plot.setLabel("left", "Power")
        self.spect_plot.setLabel("bottom", "Frequency (Hz)")
        self.spect_curve = self.spect_plot.plot(pen=et.utils.pg_pen_cycler(1))
        self.spect_smax = et.utils.SmoothMax(self.config.update_rate / 15)
        self.spect_dft_inf_line = pg.InfiniteLine(pen=et.utils.pg_pen_cycler(1, "--"))
        self.spect_plot.addItem(self.spect_dft_inf_line)
        self.spect_est_inf_line = pg.InfiniteLine(pen=et.utils.pg_pen_cycler(0, "--"))
        self.spect_plot.addItem(self.spect_est_inf_line)
        self.spect_plot.setXRange(0, 1)
        self.spect_plot.setYRange(0, 1)
        self.spect_text_item = pg.TextItem("Initiating...", anchor=(0.5, 0.5), color="k")
        self.spect_text_item.setPos(0.5, 0.5)
        self.spect_plot.addItem(self.spect_text_item)

        win.nextRow()
        self.fest_plot = win.addPlot(title="Breathing estimation history")
        self.fest_plot.setMenuEnabled(False)
        self.fest_plot.setMouseEnabled(x=False, y=False)
        self.fest_plot.hideButtons()
        self.fest_plot.showGrid(x=True, y=True)
        self.fest_plot.setLabel("left", "Frequency (Hz)")
        self.fest_plot.setLabel("bottom", "Samples")
        self.fest_plot.addLegend()
        self.fest_curve = self.fest_plot.plot(
            pen=et.utils.pg_pen_cycler(0),
            name="Breathing est.",
        )
        self.fest_dft_curve = self.fest_plot.plot(
            pen=et.utils.pg_pen_cycler(1),
            name="DFT est.",
        )
        self.fest_plot.setXRange(0, 1)
        self.fest_plot.setYRange(0, 0.5)
        self.fest_text_item = pg.TextItem(anchor=(0, 0), color="k")
        self.fest_text_item.setPos(0, 0.5)
        self.fest_plot.addItem(self.fest_text_item)

    def update(self, data):
        self.filt_phi_curve.setData(np.squeeze(data["phi_filt"]))
        self.raw_phi_curve.setData(np.squeeze(data["phi_raw"]))

        if data["init_progress"] is not None:
            self.spect_text_item.setText("Initiating: {} %".format(data["init_progress"]))
        else:
            snr = data["snr"]
            if snr == 0:
                s = "SNR: N/A | {:.0f} dB".format(10 * np.log10(data["lambda_p"]))
            else:
                fmt = "SNR: {:.0f} | {:.0f} dB"
                s = fmt.format(10 * np.log10(snr), 10 * np.log10(data["lambda_p"]))
            self.spect_text_item.setText(s)
            self.spect_text_item.setAnchor((0, 1))
            self.spect_text_item.setPos(0, 0)

            f_est = data["f_est"]
            if f_est > 0:
                s = "Latest frequency estimate: {:.2f} Hz | {:.0f} BPM".format(f_est, f_est * 60)
                self.fest_text_item.setText(s)

            self.fest_plot.enableAutoRange(x=True)
            self.spect_curve.setData(data["x_dft"], data["power_spectrum"])
            self.spect_dft_inf_line.setValue(data["f_dft_est"])
            self.spect_est_inf_line.setValue(data["f_est"])
            self.spect_plot.setYRange(0, self.spect_smax.update(data["power_spectrum"]))
            self.fest_curve.setData(np.squeeze(data["f_est_hist"]))
            self.fest_dft_curve.setData(np.squeeze(data["f_dft_est_hist"]))



import numpy as np

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et

from ._processor import ProcessingConfiguration
from .constants import EST_VEL_HISTORY_LENGTH, HALF_WAVELENGTH, NUM_SAVED_SEQUENCES


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.processing_config = processing_config

        self.sweeps_per_frame = sensor_config.sweeps_per_frame
        self.sweep_rate = session_info["sweep_rate"]
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.num_depths = self.depths.size
        self.est_update_rate = self.sweep_rate / self.sweeps_per_frame

        self.num_shown_sequences = processing_config.num_shown_sequences

        if (
            self.processing_config.processing_method
            == ProcessingConfiguration.ProcessingMethod.WELCH
        ):
            segment_length = 2 * self.sweeps_per_frame // (processing_config.num_segments + 1)
        else:
            segment_length = self.sweeps_per_frame // processing_config.num_segments

        fft_length = segment_length * processing_config.fft_oversampling_factor
        self.bin_vs = np.fft.rfftfreq(fft_length) * self.sweep_rate * HALF_WAVELENGTH
        self.dt = 1.0 / self.est_update_rate

        self.setup_is_done = False

    def setup(self, win):
        # Data plots

        self.data_plots = []
        self.data_curves = []
        for i in range(self.num_depths):
            title = "{:.0f} cm".format(100 * self.depths[i])
            plot = win.addPlot(row=0, col=i, title=title)
            plot.setMenuEnabled(False)
            plot.setMouseEnabled(x=False, y=False)
            plot.hideButtons()
            plot.showGrid(x=True, y=True)
            plot.setYRange(0, 2**16)
            plot.hideAxis("left")
            plot.hideAxis("bottom")
            plot.plot(np.arange(self.sweeps_per_frame), 2**15 * np.ones(self.sweeps_per_frame))
            curve = plot.plot(pen=et.utils.pg_pen_cycler())
            self.data_plots.append(plot)
            self.data_curves.append(curve)

        # Spectral density plot

        self.sd_plot = win.addPlot(row=1, col=0, colspan=self.num_depths)
        self.sd_plot.setMenuEnabled(False)
        self.sd_plot.setMouseEnabled(x=False, y=False)
        self.sd_plot.hideButtons()
        self.sd_plot.setLabel("left", "Normalized PSD (dB)")
        self.sd_plot.showGrid(x=True, y=True)
        self.sd_curve = self.sd_plot.plot(pen=et.utils.pg_pen_cycler())
        dashed_pen = pg.mkPen("k", width=2, style=QtCore.Qt.DashLine)
        self.sd_threshold_line = pg.InfiniteLine(angle=0, pen=dashed_pen)
        self.sd_plot.addItem(self.sd_threshold_line)

        self.smooth_max = et.utils.SmoothMax(
            self.est_update_rate,
            tau_decay=0.5,
            tau_grow=0,
            hysteresis=0.2,
        )

        # Rolling speed plot

        self.vel_plot = pg.PlotItem()
        self.vel_plot.setMenuEnabled(False)
        self.vel_plot.setMouseEnabled(x=False, y=False)
        self.vel_plot.hideButtons()
        self.vel_plot.setLabel("bottom", "Time (s)")
        self.vel_plot.showGrid(x=True, y=True)
        self.vel_plot.setXRange(-EST_VEL_HISTORY_LENGTH, 0)
        self.vel_max_line = pg.InfiniteLine(angle=0, pen=pg.mkPen("k", width=1))
        self.vel_plot.addItem(self.vel_max_line)
        self.vel_scatter = pg.ScatterPlotItem(size=8)
        self.vel_plot.addItem(self.vel_scatter)

        self.vel_html_fmt = '<span style="color:#000;font-size:24pt;">{:.1f} {}</span>'
        self.vel_text_item = pg.TextItem(anchor=(0.5, 0))
        self.vel_plot.addItem(self.vel_text_item)

        # Sequence speed plot

        self.sequences_plot = pg.PlotItem()
        self.sequences_plot.setMenuEnabled(False)
        self.sequences_plot.setMouseEnabled(x=False, y=False)
        self.sequences_plot.hideButtons()
        self.sequences_plot.setLabel("bottom", "History")
        self.sequences_plot.showGrid(y=True)
        self.sequences_plot.setXRange(-self.num_shown_sequences + 0.5, 0.5)
        tmp = np.flip(np.arange(NUM_SAVED_SEQUENCES) == 0)
        brushes = [pg.mkBrush(et.utils.color_cycler(n)) for n in tmp]
        self.bar_graph = pg.BarGraphItem(
            x=np.arange(-NUM_SAVED_SEQUENCES, 0) + 1,
            height=np.zeros(NUM_SAVED_SEQUENCES),
            width=0.8,
            brushes=brushes,
        )
        self.sequences_plot.addItem(self.bar_graph)

        self.sequences_text_item = pg.TextItem(anchor=(0.5, 0))
        self.sequences_plot.addItem(self.sequences_text_item)

        sublayout = win.addLayout(row=2, col=0, colspan=self.num_depths)
        sublayout.addItem(self.vel_plot, col=0)
        sublayout.addItem(self.sequences_plot, col=1)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        for plot in self.data_plots:
            plot.setVisible(processing_config.show_data_plot)

        self.sd_plot.setVisible(processing_config.show_sd_plot)
        self.vel_plot.setVisible(processing_config.show_vel_history_plot)

        self.unit = processing_config.shown_speed_unit
        speed_label = "Speed ({})".format(self.unit.label)
        self.sd_plot.setLabel("bottom", speed_label)
        self.vel_plot.setLabel("left", speed_label)
        self.sequences_plot.setLabel("left", speed_label)
        max_vel = self.bin_vs[-1] * self.unit.scale
        self.sd_plot.setXRange(0, max_vel)

        self.num_shown_sequences = processing_config.num_shown_sequences
        self.sequences_plot.setXRange(-self.num_shown_sequences + 0.5, 0.5)

        y_max = max_vel * 1.2
        self.vel_plot.setYRange(0, y_max)
        self.sequences_plot.setYRange(0, y_max)
        self.vel_text_item.setPos(-EST_VEL_HISTORY_LENGTH / 2, y_max)
        self.sequences_text_item.setPos(-self.num_shown_sequences / 2 + 0.5, y_max)

    def update(self, data):
        # Data plots

        for i, ys in enumerate(data["frame"].T):
            self.data_curves[i].setData(ys)

        # Spectral density plot

        psd_db = 20 * np.log10(data["nasd_temporal_max"])
        psd_threshold_db = 20 * np.log10(data["temporal_max_threshold"])
        m = self.smooth_max.update(max(2 * psd_threshold_db, np.max(psd_db)))
        self.sd_plot.setYRange(0, m)
        self.sd_curve.setData(self.bin_vs * self.unit.scale, psd_db)
        self.sd_threshold_line.setPos(psd_threshold_db)

        # Rolling speed plot

        vs = data["vel_history"] * self.unit.scale
        mask = ~np.isnan(vs)
        ts = -np.flip(np.arange(vs.size)) * self.dt
        bs = data["belongs_to_last_sequence"]
        brushes = [et.utils.pg_brush_cycler(int(b)) for b in bs[mask]]

        self.vel_scatter.setData(ts[mask], vs[mask], brush=brushes)

        v = data["vel"]
        if v:
            html = self.vel_html_fmt.format(v * self.unit.scale, self.unit.label)
            self.vel_text_item.setHtml(html)
            self.vel_text_item.show()

            self.vel_max_line.setPos(v * self.unit.scale)
            self.vel_max_line.show()
        else:
            self.vel_text_item.hide()
            self.vel_max_line.hide()

        # Sequence speed plot

        hs = data["sequence_vels"] * self.unit.scale
        self.bar_graph.setOpts(height=hs)

        if hs[-1] > 1e-3:
            html = self.vel_html_fmt.format(hs[-1], self.unit.label)
            self.sequences_text_item.setHtml(html)



from PySide6 import QtGui

import pyqtgraph as pg

import acconeer.exptool as et

from ._processor import ProcessingConfiguration


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.depth_res = session_info["step_length_m"]
        self.smooth_max = et.utils.SmoothMax(sensor_config.update_rate)

        self.setup_is_done = False

    def setup(self, win):
        num_sensors = len(self.sensor_config.sensor)

        self.ampl_plot = win.addPlot(row=0, colspan=num_sensors)
        self.ampl_plot.setMenuEnabled(False)
        self.ampl_plot.setMouseEnabled(x=False, y=False)
        self.ampl_plot.hideButtons()
        self.ampl_plot.showGrid(x=True, y=True)
        self.ampl_plot.setLabel("bottom", "Depth (m)")
        self.ampl_plot.setLabel("left", "Amplitude")
        self.ampl_plot.setXRange(*self.depths.take((0, -1)))
        self.ampl_plot.setYRange(0, 1)  # To avoid rendering bug
        self.ampl_plot.addLegend(offset=(-10, 10))

        self.ampl_curves = []
        self.bg_curves = []
        self.peak_lines = []
        for i, sensor_id in enumerate(self.sensor_config.sensor):
            legend = "Sensor {}".format(sensor_id)
            ampl_curve = self.ampl_plot.plot(pen=et.utils.pg_pen_cycler(i), name=legend)
            bg_curve = self.ampl_plot.plot(pen=et.utils.pg_pen_cycler(i, style="--"))
            color_tuple = et.utils.hex_to_rgb_tuple(et.utils.color_cycler(i))
            peak_line = pg.InfiniteLine(pen=pg.mkPen(pg.mkColor(*color_tuple, 150), width=2))
            self.ampl_plot.addItem(peak_line)
            self.ampl_curves.append(ampl_curve)
            self.bg_curves.append(bg_curve)
            self.peak_lines.append(peak_line)

        bg = pg.mkColor(0xFF, 0xFF, 0xFF, 150)
        self.peak_text = pg.TextItem(anchor=(0, 1), color="k", fill=bg)
        self.peak_text.setPos(self.depths[0], 0)
        self.peak_text.setZValue(100)
        self.ampl_plot.addItem(self.peak_text)

        rate = self.sensor_config.update_rate
        xlabel = "Sweeps" if rate is None else "Time (s)"
        x_scale = 1.0 if rate is None else 1.0 / rate
        y_scale = self.depth_res
        x_offset = -self.processing_config.history_length * x_scale
        y_offset = self.depths[0] - 0.5 * self.depth_res
        is_single_sensor = len(self.sensor_config.sensor) == 1

        self.history_plots = []
        self.history_ims = []
        for i, sensor_id in enumerate(self.sensor_config.sensor):
            title = None if is_single_sensor else "Sensor {}".format(sensor_id)
            plot = win.addPlot(row=1, col=i, title=title)
            plot.setMenuEnabled(False)
            plot.setMouseEnabled(x=False, y=False)
            plot.hideButtons()
            plot.setLabel("bottom", xlabel)
            plot.setLabel("left", "Depth (m)")
            im = pg.ImageItem(autoDownsample=True)
            im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
            im.resetTransform()
            tr = QtGui.QTransform()
            tr.translate(x_offset, y_offset)
            tr.scale(x_scale, y_scale)
            im.setTransform(tr)
            plot.addItem(im)
            self.history_plots.append(plot)
            self.history_ims.append(im)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        self.show_peaks = processing_config.show_peak_depths
        self.peak_text.setVisible(self.show_peaks)

        show_bg = processing_config.bg_mode == ProcessingConfiguration.BackgroundMode.LIMIT

        for curve in self.bg_curves:
            curve.setVisible(show_bg)

    def update(self, d):
        sweeps = d["output_data"]
        bgs = d["bg"]
        histories = d["history"]

        for i, _ in enumerate(self.sensor_config.sensor):
            self.ampl_curves[i].setData(self.depths, sweeps[i])

            if bgs is None:
                self.bg_curves[i].clear()
            else:
                self.bg_curves[i].setData(self.depths, bgs[i])

            peak = d["peak_depths"][i]
            if peak is not None and self.show_peaks:
                self.peak_lines[i].setPos(peak)
                self.peak_lines[i].show()
            else:
                self.peak_lines[i].hide()

            im = self.history_ims[i]
            history = histories[:, i]
            im.updateImage(history, levels=(0, 1.05 * history.max()))

        m = self.smooth_max.update(sweeps.max())
        self.ampl_plot.setYRange(0, m)

        # Update peak text item
        val_strs = ["-" if p is None else "{:5.3f} m".format(p) for p in d["peak_depths"]]
        z = zip(self.sensor_config.sensor, val_strs)
        t = "\n".join(["Sensor {}: {}".format(sid, v) for sid, v in z])
        self.peak_text.setText(t)



import numpy as np

from PySide6 import QtGui

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.processing_config = processing_config

        self.f = sensor_config.update_rate
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.downsampling_factor = sensor_config.downsampling_factor
        self.step_length = session_info["step_length_m"]

        self.td_smooth_lims = et.utils.SmoothLimits()
        self.collapsed_smooth_max = et.utils.SmoothMax(
            tau_grow=0.1,
        )

        self.setup_is_done = False

    def setup(self, win):
        self.td_plot = win.addPlot(row=0, col=0, title="PSD input data")
        self.td_plot.setMenuEnabled(False)
        self.td_plot.setMouseEnabled(x=False, y=False)
        self.td_plot.hideButtons()
        self.td_plot.addLegend()
        self.td_curves = []
        for i, depth in enumerate(self.depths):
            name = "{:.0f} cm".format(depth * 100)
            curve = self.td_plot.plot(pen=et.utils.pg_pen_cycler(i), name=name)
            self.td_curves.append(curve)

        self.collapsed_plot = win.addPlot(row=1, col=0, title="Collapsed sqrt(PSD)")
        self.collapsed_plot.setMenuEnabled(False)
        self.collapsed_plot.setMouseEnabled(x=False, y=False)
        self.collapsed_plot.hideButtons()
        self.collapsed_plot.setXRange(0, 1)
        self.collapsed_curve = self.collapsed_plot.plot(pen=et.utils.pg_pen_cycler())
        self.collapsed_vline = pg.InfiniteLine(pen=et.utils.pg_pen_cycler())
        self.collapsed_vline.hide()
        self.collapsed_plot.addItem(self.collapsed_vline)

        bg = pg.mkColor(0xFF, 0xFF, 0xFF, 150)
        self.collapsed_text = pg.TextItem(anchor=(0, 1), color="k", fill=bg)
        self.collapsed_text.setPos(0, 0)
        self.collapsed_text.setZValue(100)
        self.collapsed_plot.addItem(self.collapsed_text, ignoreBounds=True)

        self.collapsed_history_plot = win.addPlot(
            row=2, col=0, title="Collapsed sqrt(PSD) history"
        )
        self.collapsed_history_plot.setMenuEnabled(False)
        self.collapsed_history_plot.setMouseEnabled(x=False, y=False)
        self.collapsed_history_plot.hideButtons()
        self.collapsed_history_im = pg.ImageItem()
        self.collapsed_history_im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
        self.collapsed_history_plot.addItem(self.collapsed_history_im)

        self.dw_plot = win.addPlot(row=3, col=0, title="Depthwise PSD")
        self.dw_plot.setMenuEnabled(False)
        self.dw_plot.setMouseEnabled(x=False, y=False)
        self.dw_plot.hideButtons()
        self.dw_im = pg.ImageItem()
        self.dw_im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
        self.dw_plot.addItem(self.dw_im)
        self.dw_plot.setLabel("bottom", "Depth (cm)")
        self.dw_plot.getAxis("bottom").setTickSpacing(6 * self.downsampling_factor, 6)

        self.setup_is_done = True
        self.update_processing_config()

    def update_processing_config(self, processing_config=None):
        if processing_config is None:
            processing_config = self.processing_config
        else:
            self.processing_config = processing_config

        if not self.setup_is_done:
            return

        self.td_plot.setVisible(processing_config.show_time_domain)
        self.collapsed_history_plot.setVisible(processing_config.show_spect_history)
        self.dw_plot.setVisible(processing_config.show_depthwise_spect)

        lbl = "Time (s)" if self.f else "Frame index"
        self.td_plot.setLabel("bottom", lbl)

        lbl = "Frequency (Hz)" if self.f else "Frequency bin"
        self.collapsed_plot.setLabel("bottom", lbl)

        for plot in [self.dw_plot, self.collapsed_history_plot]:
            plot.setLabel("left", lbl)

        if self.f:
            f_res = self.f / self.processing_config._window_size
        else:
            f_res = 1

        tr = QtGui.QTransform()
        tr.translate(100 * (self.depths[0] - self.step_length / 2), 0)
        tr.scale(100 * self.step_length, f_res)
        self.dw_im.setTransform(tr)

        self.collapsed_history_im.resetTransform()
        tr.translate(0, f_res / 2)
        tr.scale(-1, f_res)
        self.collapsed_history_im.setTransform(tr)

    def update(self, d):
        x = d["ts"]
        for i, curve in enumerate(self.td_curves):
            y = d["sweep_history"][:, i]
            curve.setData(x, y)

        r = self.td_smooth_lims.update(d["sweep_history"])
        self.td_plot.setYRange(*r)

        x = d["fs"]
        y = d["collapsed_asd"]
        self.collapsed_curve.setData(x, y)
        m = self.collapsed_smooth_max.update(y)
        self.collapsed_plot.setXRange(0, x[-1])
        self.collapsed_plot.setYRange(0, m)

        f_max = x[y.argmax()]
        self.collapsed_vline.setPos(f_max)
        self.collapsed_vline.show()

        if self.f:
            s = "Peak: {:5.1f} Hz".format(f_max)
        else:
            s = "Peak: {:3.0f}".format(f_max)
        self.collapsed_text.setText(s)

        im = self.collapsed_history_im
        y = d["collapsed_asd_history"]
        im.updateImage(y[::-1], levels=(0, 1.05 * y.max()))

        y = d["dw_asd"]
        m = max(1, np.max(y)) * 1.05
        self.dw_im.updateImage(y, levels=(0, m))



from enum import Enum

import numpy as np

import acconeer.exptool as et

from .calibration import EnvelopeCalibration


def get_sensor_config():
    config = et.a111.EnvelopeServiceConfig()
    config.range_interval = [0.2, 0.8]
    config.hw_accelerated_average_samples = 15
    config.update_rate = 30
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1

    class BackgroundMode(Enum):
        SUBTRACT = "Subtract"
        LIMIT = "Limit"

    show_peak_depths = et.configbase.BoolParameter(
        label="Show peak distances",
        default_value=True,
        updateable=True,
        order=-10,
    )

    bg_buffer_length = et.configbase.IntParameter(
        default_value=50,
        limits=(1, 200),
        label="Background buffer length",
        order=0,
    )

    bg_mode = et.configbase.EnumParameter(
        label="Background mode",
        default_value=BackgroundMode.SUBTRACT,
        enum=BackgroundMode,
        updateable=True,
        order=20,
    )

    history_length = et.configbase.IntParameter(
        default_value=100,
        limits=(10, 1000),
        label="History length",
        order=30,
    )


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.processing_config = processing_config

        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        num_depths = self.depths.size
        num_sensors = len(sensor_config.sensor)

        buffer_length = self.processing_config.bg_buffer_length
        self.bg_buffer = np.zeros([buffer_length, num_sensors, num_depths])

        history_length = self.processing_config.history_length
        self.history = np.zeros([history_length, num_sensors, num_depths])

        self.data_index = 0
        self.calibration = calibration

    def process(self, data, data_info):
        new_calibration = None
        bg = None
        output_data = data

        if self.calibration is None:
            if self.data_index < self.bg_buffer.shape[0]:
                self.bg_buffer[self.data_index] = data
            if self.data_index == self.bg_buffer.shape[0] - 1:
                new_calibration = EnvelopeCalibration(self.bg_buffer.mean(axis=0))
        else:
            if self.processing_config.bg_mode == ProcessingConfiguration.BackgroundMode.SUBTRACT:
                output_data = np.maximum(0, data - self.calibration.background)
            else:
                output_data = np.maximum(data, self.calibration.background)

            bg = self.calibration.background

        self.history = np.roll(self.history, -1, axis=0)
        self.history[-1] = output_data

        peak_ampls = [np.max(sweep) for sweep in output_data]
        peak_depths = [self.depths[np.argmax(sweep)] for sweep in output_data]
        filtered_peak_depths = [d if a > 200 else None for d, a in zip(peak_depths, peak_ampls)]

        output = {
            "output_data": output_data,
            "bg": bg,
            "history": self.history,
            "peak_depths": filtered_peak_depths,
        }
        if new_calibration is not None:
            output["new_calibration"] = new_calibration

        self.data_index += 1

        return output

    def update_calibration(self, new_calibration: EnvelopeCalibration):
        self.calibration = new_calibration
        self.data_index = 0



import acconeer.exptool as et
from acconeer.exptool.a111.algo import ModuleInfo


def main(module_info: ModuleInfo):
    arg_parse_kwargs = {} if module_info.multi_sensor else dict(num_sens=1)
    args = et.a111.ExampleArgumentParser(**arg_parse_kwargs).parse_args()
    et.utils.config_logging(args)

    client = et.a111.Client(**et.a111.get_client_args(args))

    client.squeeze = not module_info.multi_sensor

    sensor_config = module_info.sensor_config_class()
    processing_config = module_info.processing_config_class()
    sensor_config.sensor = args.sensors

    session_info = client.setup_session(sensor_config)

    pg_updater = module_info.pg_updater(sensor_config, processing_config, session_info)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    processor = module_info.processor(sensor_config, processing_config, session_info)

    while not interrupt_handler.got_signal:
        info, sweep = client.get_next()
        processed_data = processor.process(sweep, info)

        if processed_data is not None:
            if hasattr(processor, "update_calibration"):
                new_calibration = processed_data.get("new_calibration")
                if new_calibration is not None:
                    processor.update_calibration(new_calibration)

            try:
                pg_process.put_data(processed_data)
            except et.PGProccessDiedException:
                break

    print("Disconnecting...")
    pg_process.close()
    client.disconnect()



import numpy as np
import scipy.signal
import scipy.stats

import acconeer.exptool as et

from .calibration import EnvelopeCalibration


SUB_MEAN_PIECE_LENGTH = 250


def get_sensor_config():
    config = et.a111.EnvelopeServiceConfig()
    config.range_interval = [0.0, 0.4]
    config.update_rate = 20
    config.gain = 0.0
    config.profile = config.Profile.PROFILE_1
    config.hw_accelerated_average_samples = 40
    config.running_average_factor = 0  # Use averaging in detector instead of in API
    config.noise_level_normalization = True
    config.maximize_signal_attenuation = False
    config.downsampling_factor = 1

    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1

    precision = et.configbase.IntParameter(
        default_value=150,
        limits=(1, 2000),
        label="Precision",
        updateable=False,
        order=0,
        help=(
            "Number of possible output values. "
            "High number yields slower performance and finer grain of output."
        ),
    )

    bg_buffer_length = et.configbase.IntParameter(
        default_value=50,
        limits=(1, 200),
        label="Background buffer length",
        order=10,
    )

    history_length_s = et.configbase.IntParameter(
        default_value=10,
        unit="s",
        limits=(10, 1000),
        label="History length",
        order=20,
    )

    smoothing_time_const = et.configbase.FloatParameter(
        label="Smoothing time const.",
        unit="s",
        default_value=0.5,
        limits=(0.01, 30),
        logscale=True,
        updateable=True,
        order=30,
        help=("Time constant for the smoothing filter of the normalized raw data."),
    )

    score_threshold = et.configbase.FloatParameter(
        label="Confidence threshold to return a guess.",
        default_value=0.9,
        limits=(0.5, 1.0),
        logscale=True,
        updateable=True,
        order=40,
        help=("Threshold for the similarity measurement to return a value."),
    )

    def check_sensor_config(self, sensor_config):
        alerts = {
            "processing": [],
            "sensor": [],
        }

        if (
            sensor_config.profile != sensor_config.Profile.PROFILE_1
            and sensor_config.profile != sensor_config.Profile.PROFILE_2
        ):
            alerts["sensor"].append(et.configbase.Error("profile", "Only profile 1 or 2."))

        if sensor_config.update_rate is None:
            alerts["sensor"].append(et.configbase.Error("update_rate", "Must be set"))

        if (
            sensor_config.profile == sensor_config.Profile.PROFILE_1
            and sensor_config.range_interval[1] > 0.6
        ) or (
            sensor_config.profile == sensor_config.Profile.PROFILE_2
            and sensor_config.range_interval[1] > 0.8
        ):
            alerts["sensor"].append(
                et.configbase.Warning("range_interval", "Long range might yield poor results.")
            )

        if sensor_config.downsampling_factor > 1:
            alerts["sensor"].append(
                et.configbase.Error("downsampling_factor", "Only downsampling factor 1.")
            )

        return alerts


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.session_info = session_info
        self.sensor_config = sensor_config
        self.processing_config = processing_config

        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.num_depths = self.depths.size
        self.num_sensors = len(sensor_config.sensor)

        self.score_threshold = processing_config.score_threshold

        self.calibration = calibration
        buffer_length = self.processing_config.bg_buffer_length
        self.bg_buffer = np.zeros([buffer_length, self.num_depths])

        self.range_start = self.session_info["range_start_m"]
        self.range_length = self.session_info["range_length_m"]

        self.data_index = 0

        tc = processing_config.smoothing_time_const
        self.fs = sensor_config.update_rate
        alpha = self.tc_to_sf(tc, self.fs)

        self.history_length_s = processing_config.history_length_s

        self.smooth_const = alpha
        self.smooth_val = np.zeros(self.num_depths)

        self.depths = et.a111.get_range_depths(sensor_config, session_info)

        self.scale_array = self.depths / max(self.depths)

        self.peak_width = self.get_peak_width(sensor_config)
        self.masks = self.make_mask_list(processing_config.precision)

        self.guess_history = np.zeros(self.history_length_s)
        self.guess_index_history = np.zeros(self.history_length_s)

    def get_peak_width(self, sensor_config):
        profile = sensor_config.profile
        if profile == sensor_config.Profile.PROFILE_1:
            width = 20
        elif profile == sensor_config.Profile.PROFILE_2:
            width = 40
        else:
            raise ValueError("Illegal profile ({}) for this detector".format(profile))
        return width

    def update_processing_config(self, processing_config):
        tc = processing_config.smoothing_time_const
        alpha = self.tc_to_sf(tc, self.fs)
        self.smooth_const = alpha
        self.score_threshold = processing_config.score_threshold

    def tc_to_sf(self, tc, fs):  # time constant to smoothing factor conversion
        if tc <= 0.0:
            return 0.0

        return np.exp(-1.0 / (tc * fs))

    def sub_mean(self, data):
        """
        Estimate of noise floor.
        Decent estimate even with object in range.
        Uses subsampling.
        """
        piece_len = SUB_MEAN_PIECE_LENGTH
        min_mean = data.max()
        min_dev = data.max()
        start = 0
        end = min(piece_len, data.shape[0])
        for start in range(end):
            d_slice = data[start::piece_len]
            sub_mean = np.mean(d_slice)
            dev = np.std(d_slice)

            if sub_mean < min_mean:
                min_mean = sub_mean
                min_dev = dev

        return min_mean + min_dev

    def normalize(self, data):
        div = np.amax(data)
        if div != 0.0:
            data = data / div
        return data

    def get_distance(self, index):
        dist = self.range_start + (index / self.num_depths) * self.range_length
        return dist

    def get_index(self, distance):
        index = self.num_depths * (distance - self.range_start) / self.range_length
        return int(round(index))

    def make_peak(self, index, width):
        x = np.arange(self.num_depths)
        vals = scipy.stats.norm.pdf(x, index, width)
        return vals

    def calculate_mask(self, distance):
        mask = np.zeros(self.num_depths)

        n_peaks = 3

        for i in range(n_peaks):
            dist_to_use = (i + 1) * distance
            index = self.get_index(dist_to_use)
            peak = self.make_peak(index, self.peak_width) * (
                (2 / 3) ** i
            )  # The 2/3 is empirically chosen to best match the behaviour in the close range.
            mask = mask + peak

        mask = mask / max(mask)
        mask[mask < 0.001] = 0.0
        return mask

    def make_mask_list(self, n_points):

        mask_dists = np.linspace(self.range_start, self.range_start + self.range_length, n_points)

        masks = [[self.calculate_mask(r), r] for r in mask_dists]

        return masks

    def process(self, data, data_info):

        new_calibration = None

        if self.calibration is None:
            if self.data_index < self.bg_buffer.shape[0]:
                self.bg_buffer[self.data_index] = data
            if self.data_index == self.bg_buffer.shape[0] - 1:
                new_calibration = EnvelopeCalibration(self.bg_buffer.mean(axis=0))
            mean = self.sub_mean(data)
            data = np.maximum(0, data - mean)
        else:
            data = np.maximum(0, data - self.calibration.background)

        plot_data = {}
        plot_data["smooth_data"] = np.zeros(data.shape[0])
        plot_data["index"] = self.data_index

        data = data * self.scale_array  # Scale by distance

        data = self.normalize(data)  # Allow for using a set threshold.

        # Necessary since we have normalized noise.
        self.smooth_val = self.smooth_val * self.smooth_const + (1.0 - self.smooth_const) * data

        best_score = -1.0
        guess = 0.0

        for mask, dist in self.masks:
            match = np.minimum(mask, self.smooth_val)
            overlay = np.maximum(mask, self.smooth_val)
            miss = overlay - match
            score = 1.0 - (sum(miss) / self.num_depths)

            if score > best_score:
                best_score = score
                guess = dist

        if best_score < self.score_threshold:  # test if we have a reasonable guess.
            guess = 0.0

        plot_data["best_guess"] = guess

        if self.data_index % self.fs == 0:
            self.guess_history = np.roll(self.guess_history, -1)
            self.guess_history[-1] = guess
            self.guess_index_history = np.roll(self.guess_index_history, -1)
            self.guess_index_history[-1] = self.data_index

        plot_data["guess_hist"] = self.guess_history
        plot_data["guess_hist_idx"] = self.guess_index_history
        plot_data["smooth_data"] = self.smooth_val

        if new_calibration is not None:
            plot_data["new_calibration"] = new_calibration

        self.data_index += 1

        return plot_data

    def update_calibration(self, new_calibration: EnvelopeCalibration):
        self.calibration = new_calibration
        self.data_index = 1
        self.guess_history = np.zeros(self.history_length_s)
        self.guess_index_history = np.zeros(self.history_length_s)



import configparser
import fnmatch
import subprocess
import sys


def main():
    config = configparser.ConfigParser()
    config.read("setup.cfg")
    section = "check_whitespace"

    ignore_files = config.get(section, "ignore_files", fallback="")
    ignore_files = [s.strip() for s in ignore_files.split(",")]
    ignore_files = [s for s in ignore_files if s]

    p = subprocess.run(
        ["git", "ls-files"],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        check=True,
    )

    filenames = p.stdout.decode().splitlines()

    failed = False
    for filename in filenames:
        if any(True for p in ignore_files if fnmatch.fnmatch(filename, p)):
            continue

        try:
            check(filename)
        except UnicodeDecodeError:
            pass
        except Exception as e:
            failed = True
            print("{}: {}".format(filename, e))

    return failed


def check(filename):
    last_is_empty = False

    with open(filename, "r", newline="") as f:
        for i, line in enumerate(f):
            if line.endswith("\r") or line.endswith("\r\n"):
                raise Exception("Carriage returns are not allowed")

            if not line.endswith("\n"):
                raise Exception("Missing newline at end of file")

            try:
                last_char = line[-2]
            except IndexError:
                last_is_empty = True
            else:
                if last_char.isspace():
                    raise Exception("Trailing whitespace at end of line {}".format(i + 1))

                last_is_empty = False

    if last_is_empty:
        raise Exception("Trailing newlines at end of file")


if __name__ == "__main__":
    sys.exit(int(main()))



import numpy as np
from numpy import pi
from scipy.signal import butter, sosfilt

import acconeer.exptool as et


def get_sensor_config():
    config = et.a111.IQServiceConfig()
    config.range_interval = [0.3, 0.8]
    config.update_rate = 80
    config.gain = 0.5
    config.repetition_mode = et.a111.IQServiceConfig.RepetitionMode.SENSOR_DRIVEN
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1

    hist_plot_len = et.configbase.FloatParameter(
        label="Plot length",
        unit="s",
        default_value=10,
        limits=(1, 30),
        decimals=0,
    )


class Processor:
    peak_hist_len = 600

    phase_weights_alpha = 0.9
    peak_loc_alpha = 0.95
    sweep_alpha = 0.7
    env_alpha = 0.95

    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.config = sensor_config

        assert sensor_config.update_rate is not None

        self.f = sensor_config.update_rate
        self.hist_plot_len = int(round(processing_config.hist_plot_len * self.f))
        self.breath_hist_len = max(2000, self.hist_plot_len)

        self.peak_history = np.zeros(self.peak_hist_len, dtype="complex")
        self.movement_history = np.zeros(self.peak_hist_len, dtype="float")
        self.breath_history = np.zeros(self.breath_hist_len, dtype="float")
        self.pulse_history = np.zeros(self.hist_plot_len, dtype="float")

        self.breath_sos = np.concatenate(butter(2, 2 * 0.3 / self.f))
        self.breath_zi = np.zeros((1, 2))
        self.pulse_sos = np.concatenate(butter(2, 2 * 5 / self.f))
        self.pulse_zi = np.zeros((1, 2))

        self.last_lp_sweep = None
        self.lp_phase_weights = None
        self.lp_sweep = None
        self.lp_peak_loc = 0

        self.sweep_index = 0

    def process(self, data, data_info):
        sweep = data

        if self.sweep_index == 0:
            self.lp_sweep = np.array(sweep)
            self.lp_env = np.abs(sweep)
            self.lp_peak_loc = np.argmax(self.lp_env)

            out_data = None
        else:
            self.lp_sweep = self.lp(sweep, self.lp_sweep, self.sweep_alpha)
            env = np.abs(self.lp_sweep)
            self.lp_env = self.lp(env, self.lp_env, self.env_alpha)
            peak_loc = np.argmax(self.lp_env)
            self.lp_peak_loc = self.lp(peak_loc, self.lp_peak_loc, self.peak_loc_alpha)

            peak_idx = int(round(self.lp_peak_loc))
            peak = np.mean(self.lp_sweep[peak_idx - 50 : peak_idx + 50])
            self.push(peak, self.peak_history)

            delta = self.lp_sweep * np.conj(self.last_lp_sweep)

            phase_weights = np.imag(delta)
            if self.lp_phase_weights is None:
                self.lp_phase_weights = phase_weights
            else:
                self.lp_phase_weights = self.lp(
                    phase_weights,
                    self.lp_phase_weights,
                    self.phase_weights_alpha,
                )

            weights = np.abs(self.lp_phase_weights) * env

            delta_dist = np.dot(weights, np.angle(delta))
            delta_dist *= 2.5 / (2.0 * pi * sum(weights + 0.00001))

            y = self.movement_history[0] + delta_dist
            self.push(y, self.movement_history)

            y_breath, self.breath_zi = sosfilt(self.breath_sos, np.array([y]), zi=self.breath_zi)
            self.push(y_breath, self.breath_history)

            y_pulse, self.pulse_zi = sosfilt(self.pulse_sos, np.array([y]), zi=self.pulse_zi)
            self.push(y_pulse, self.pulse_history)

            maxs = self.find_peaks(self.breath_history, 100)
            mins = self.find_peaks(-self.breath_history, 100)
            max_idx = 0
            min_idx = 0
            inhale_time = None
            exhale_time = None
            inhale_dist = 0
            exhale_dist = 0
            exhale = None
            first_peak = None
            while not (inhale_time and exhale_time):
                if not (min_idx < mins.shape[0] and max_idx < maxs.shape[0]):
                    break

                if maxs[max_idx, 0] < mins[min_idx, 0]:
                    exhale_dist = mins[min_idx, 1] + maxs[max_idx, 1]
                    if exhale_dist > 1 and exhale_dist < 20:
                        exhale_time = mins[min_idx, 0] - maxs[max_idx, 0]
                        if first_peak is None:
                            first_peak = maxs[max_idx, 0]
                        exhale = True
                    max_idx += 1
                else:
                    inhale_dist = mins[min_idx, 1] + maxs[max_idx, 1]
                    if inhale_dist > 1 and inhale_dist < 20:
                        inhale_time = maxs[max_idx, 0] - mins[min_idx, 0]
                        exhale = False
                        if first_peak is None:
                            first_peak = mins[min_idx, 0]
                    min_idx += 1

            breathing = False
            if inhale_time is not None and exhale_time is not None:
                bpm = 60.0 / ((inhale_time + exhale_time) / self.f)
                symmetry = (inhale_dist - exhale_dist) / (inhale_dist + exhale_dist)
                first_peak_rel = first_peak / (inhale_time + exhale_time)
                if 3 < bpm < 30 and abs(symmetry) < 0.6 and first_peak_rel < 0.7:
                    breathing = True

            if breathing:
                bstr = "Exhaling" if exhale else "Inhaling"
                bpm_text = "{}, BPM {:0.1f}, depth {:0.1f} mm".format(bstr, bpm, inhale_dist)
            else:
                bpm_text = None

            # Make an explicit copy, otherwise flip will not return a new object
            breath_hist_plot = self.breath_history[: self.hist_plot_len]
            breath_hist_plot = np.array(np.flip(breath_hist_plot, axis=0))
            breath_hist_plot -= (np.max(breath_hist_plot) + np.min(breath_hist_plot)) * 0.5

            zoom_hist_plot = self.pulse_history[: self.hist_plot_len // 2]
            zoom_hist_plot = np.array(np.flip(zoom_hist_plot, axis=0))
            zoom_hist_plot -= (max(zoom_hist_plot) + min(zoom_hist_plot)) * 0.5

            out_data = {
                "peak_hist": self.peak_history[:100],
                "peak_std_mm": 2.5 * np.std(np.unwrap(np.angle(self.peak_history))) / (2.0 * pi),
                "env_ampl": abs(self.lp_sweep),
                "env_delta": self.lp_phase_weights,
                "peak_idx": peak_idx,
                "breathing_history": breath_hist_plot,
                "breathing_text": bpm_text,
                "zoom_hist": zoom_hist_plot,
            }

        self.last_lp_sweep = self.lp_sweep
        self.sweep_index += 1
        return out_data

    def lp(self, new, state, alpha):
        return alpha * state + (1 - alpha) * new

    def push(self, val, arr):
        res = np.empty_like(arr)
        res[0] = val
        res[1:] = arr[:-1]
        arr[...] = res

    def find_peaks(self, env, width):
        n = len(env)
        peaks = np.zeros((0, 2))
        for idx in range(0, n, width):
            mi = np.argmax(env[idx : min(idx + width, n)]) + idx
            mi2 = np.argmax(env[max(mi - width, 0) : min(mi + width, n)])
            mi2 += max(mi - width, 0)
            if mi == mi2 and (0 < mi < n - 1):
                peaks = np.concatenate((peaks, np.array([[mi, env[mi]]])), axis=0)
        return peaks
